# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""This module provides classes that specify how RESTful resources are handled.

THE MOST IMPORTANT THING TO KNOW ABOUT THIS API IS THAT YOU CAN DIRECTLY INFER
REST URIs FROM PYTHON EXPRESSIONS, AND VICE VERSA.

Examples:

 * Expression:     bigip = ManagementRoot('a', 'b', 'c')
 * URI Returned:   https://a/mgmt/

 * Expression:     bigip.tm.ltm
 * URI Returned:   https://a/mgmt/tm/ltm/

 * Expression:     pools1 = bigip.tm.ltm.pools
 * URI Returned:   https://a/mgmt/tm/ltm/pool

 * Expression:     pool_a = pools1.create(partition="Common", name="foo")
 * URI Returned:   https://a/mgmt/tm/ltm/pool/~Common~foo

There are different types of resources published by the BIG-IP® REST Server,
they are represented by the classes in this module.

We refer to a server-provided resource as a "service".  Thus far all URI
referenced resources are "services" in this sense.

We use methods named Create, Refresh, Update, Load, Modify, and Delete to
manipulate BIG-IP® device services.

Methods:

  * create -- uses HTTP POST, creates a new resource and with its own URI on
    the device
  * refresh -- uses HTTP GET, obtains the state of a device resource, and sets
    the representing Python Resource Object tracks device state via its attrs
  * update -- uses HTTP PUT, submits a new configuration to the device resource
     and sets the Resource attrs to the state the device reports
  * load -- uses HTTP GET, obtains the state of an existing resource on the
    device and sets the Resource attrs to that state
  * modify -- uses HTTP PATCH to selectively modify named resources submitted
    as keyword arguments
  * delete -- uses HTTP DELETE, removes the resource from the device, and sets
    self.__dict__ to {'deleted': True}

Available Classes:
  * PathElement -- the most fundamental class it represent URI elements that
    serve only as place-holders.  All other Resources inherit from
    PathElement, though the inheritance may be indirect. PathElement provides
    a constructor to match its call in LazyAttributeMixin.__getattr__. The
    expected behavior is that all resource subclasses depend on this
    constructor to correctly set their self._meta_data['uri'].  See
    _set_meta_data_uri for the logic underlying self._meta_data['uri']
    construction.
  * ResourceBase -- only `refresh` is generally supported in all resource
    types, this class provides `refresh`. ResourceBase objects are usually
    instantiated via setting lazy attributes.
    All ResourceBase objects (except BIG-IPs) have a container (BIG-IPs
    contain themselves).  The container is the object the ResourceBase is an
    attribute of.
  * OrganizingCollection -- These resources support lists of "reference"
    "links". These are json blobs without a Python class representation.
    Example URI_path:  /mgmt/tm/ltm/
  * Collection -- These resources support lists of ResourceBase Objects.
    Example URI_path:  /mgmt/tm/ltm/nat
  * Resource -- These resources are the only resources that support
    `create`, `update`, and `delete` operations.  Because they support HTTP
    post (via _create) they uniquely depend on 2 uri's, a uri that supports
    the creating post, and the returned uri of the newly created resource.
    Example URI_path:  /mgmt/tm/ltm/nat/~Common~testnat1
  * UnnamedResource -- Some resources correspond to URIs that do not have
    unique names, therefore the class does _not_ support create-or-delete,
    and supports a customized 'load' that doesn't require name/partition
    parameters.
"""
try:
    from collections import OrderedDict
except ImportError:
    try:
        from ordereddict import OrderedDict
    except ImportError as exc:
        message = ("Maybe you're using Python < 2.7 and do not have the "
                   "orderreddict external dependency installed.")
        raise exc(message)
import copy
import keyword
import re
import time
import tokenize
try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse

from f5.bigip.mixins import LazyAttributeMixin
from f5.bigip.mixins import ToDictMixin
from f5.sdk_exception import AttemptedMutationOfReadOnly
from f5.sdk_exception import BooleansToReduceHaveSameValue
from f5.sdk_exception import DeviceProvidesIncompatibleKey
from f5.sdk_exception import ExclusiveAttributesPresent
from f5.sdk_exception import GenerationMismatch
from f5.sdk_exception import InvalidForceType
from f5.sdk_exception import InvalidResource
from f5.sdk_exception import KindTypeMismatch
from f5.sdk_exception import MissingRequiredCommandParameter
from f5.sdk_exception import MissingRequiredCreationParameter
from f5.sdk_exception import MissingRequiredReadParameter
from f5.sdk_exception import RequestParamKwargCollision
from f5.sdk_exception import UnregisteredKind
from f5.sdk_exception import UnsupportedMethod
from f5.sdk_exception import UnsupportedOperation
from f5.sdk_exception import URICreationCollision
from icontrol.exceptions import iControlUnexpectedHTTPError
from requests.exceptions import ConnectionError
from requests.exceptions import HTTPError
from six import iteritems
from six import iterkeys
from six import itervalues


def _missing_required_parameters(rqset, **kwargs):
    """Helper function to do operation on sets.

    Checks for any missing required parameters.
    Returns non-empty or empty list. With empty
    list being False.

    ::returns list
    """
    key_set = set(list(iterkeys(kwargs)))
    required_minus_received = rqset - key_set
    if required_minus_received != set():
        return list(required_minus_received)


class PathElement(LazyAttributeMixin):
    """Base class to represent a URI path element that does not contain data.

    The BIG-IP® iControl REST API has URIs that are made up of path components
    that do not return data when they are queried.  This class represents
    those elements and does not support any of the CURDLE methods that
    the other objects do.
    """
    def __init__(self, container):
        self._meta_data = {
            'container': container,
            'bigip': container._meta_data['bigip'],
            'icr_session': container._meta_data['icr_session'],
            'icontrol_version': container._meta_data['icontrol_version']
        }
        self._set_meta_data_uri()
        # Supported versions for each class will be defined here.
        # List can be modified downstream in each sub-class
        self._meta_data['minimum_version'] = '11.5.0'
        # Commands you can run on a resource or collection, we define it here
        self._meta_data['allowed_commands'] = []
        # Define required command parameters
        self._meta_data['required_command_parameters'] = set()
        # You can't have more than one of the attributes in any of these sets.
        self._meta_data['exclusive_attributes'] = []
        # As some objects do not support Stats we need this setting,
        # by default Stats are enabled on all endpoints. Override in
        # subclass as required
        self._meta_data['object_has_stats'] = True
        # Some endpoints have mandatory parameters that can be one of x,
        # this means that all or at least 1 need to be present during create
        self._meta_data['minimum_additional_parameters'] = set()

    def _set_meta_data_uri(self):
        base_uri = self._get_base_uri()
        endpoint = base_uri.replace('_', '-')
        final_uri = self._build_final_uri(endpoint)
        self._meta_data['uri'] = final_uri

    def _get_base_uri(self):
        if isinstance(self, Collection):
            return self._format_collection_name()
        return self._format_resource_name()

    def _build_final_uri(self, endpoint):
        return self._meta_data['container']._meta_data['uri'] + endpoint + '/'

    def _format_collection_name(self):
        """Formats a name from Collection format

        Collections are of two name formats based on their actual URI
        representation in the REST service.

        1. For cases where the actual URI of a collection is singular, for
           example,

               /mgmt/tm/ltm/node

           The name of the collection, as exposed to the user, will be made
           plural. For example,

               mgmt.tm.ltm.nodes

        2. For cases where the actual URI of a collection is plural, for
           example,

               /mgmt/cm/shared/licensing/pools/

           The name of the collection, as exposed to the user, will remain
           plural, but will have an `_s` appended to it. For example,

               mgmt.cm.shared.licensing.pools_s

        This method is responsible for undoing the user provided plurality.
        It ensures that the URI that is being sent to the REST service is
        correctly plural, or plural plus.

        Returns:
            A string representation of the user formatted Collection with its
            plurality identifier removed appropriately.
        """
        base_uri = self._format_resource_name()
        if base_uri[-2:] == '_s':
            endind = 2
        else:
            endind = 1
        return base_uri[:-endind]

    def _format_resource_name(self):
        """Formats the name of a Resource

        When the names of Resources are used to create URIs, their names are
        automatically cast to lowercase to match the actual URI expected by
        the REST service.

        There are certain URIs, however, that must not be cast in this way.

        For these URIs, this method should be overloaded in the Resource
        class itself to specify how to handle the Resource name.

        An example of this difference in behavior, refer to the BIG-IQ or
        iWorkflow APIs such as,

            /mgmt/shared/resolver/device-groups/cm-bigip-allBigIpDevices/

        Returns:
            A string representation of the Resource as it should be
            represented when contructing the final URI used to reach that
            Resource.
        """
        return self.__class__.__name__.lower()

    def _check_command_parameters(self, **kwargs):
        """Params given to exec_cmd should satisfy required params.

        :params: kwargs
        :raises: MissingRequiredCommandParameter
        """
        rset = self._meta_data['required_command_parameters']
        check = _missing_required_parameters(rset, **kwargs)
        if check:
            error_message = 'Missing required params: %s' % check
            raise MissingRequiredCommandParameter(error_message)

    def _check_force_arg(self, force):
        if not isinstance(force, bool):
            raise InvalidForceType("force parameter must be type bool")
        return force

    def _check_generation(self):
        """Check that the generation on the BIG-IP® matches the object

        This will do a get to the objects URI and check that the generation
        returned in the JSON matches the one the object currently has.  If it
        does not it will raise the `GenerationMismatch` exception.
        """
        session = self._meta_data['bigip']._meta_data['icr_session']
        response = session.get(self._meta_data['uri'])
        current_gen = response.json().get('generation', None)
        if current_gen is not None and current_gen != self.generation:
            error_message = ("The generation of the object on the BigIP "
                             + "(" + str(current_gen) + ")"
                             + " does not match the current object"
                             + "(" + str(self.generation) + ")")
            raise GenerationMismatch(error_message)

    def _handle_requests_params(self, kwargs):
        """Validate parameters that will be passed to the requests verbs.

        This method validates that there is no conflict in the names of the
        requests_params passed to the function and the other kwargs.  It also
        ensures that the required request parameters for the object are
        added to the request params that are passed into the verbs.  An
        example of the latter is ensuring that a certain version of the API
        is always called to add 'ver=11.6.0' to the url.
        """
        requests_params = kwargs.pop('requests_params', {})
        for param in requests_params:
            if param in kwargs:
                error_message = 'Requests Parameter %r collides with a load'\
                    ' parameter of the same name.' % param
                raise RequestParamKwargCollision(error_message)
        # If we have an icontrol version we need to add 'ver' to params
        if self._meta_data['icontrol_version']:
            params = requests_params.pop('params', {})
            params.update({'ver': self._meta_data['icontrol_version']})
            requests_params.update({'params': params})
        return requests_params

    def _check_exclusive_parameters(self, **kwargs):
        """Check for mutually exclusive attributes in kwargs.

        :raises ExclusiveAttributesPresent
        """
        if len(self._meta_data['exclusive_attributes']) > 0:
            attr_set = set(list(iterkeys(kwargs)))
            ex_set = set(self._meta_data['exclusive_attributes'][0])
            common_set = sorted(attr_set.intersection(ex_set))
            if len(common_set) > 1:
                cset = ', '.join(common_set)
                error = 'Mutually exclusive arguments submitted. ' \
                        'The following arguments cannot be set ' \
                        'together: "%s".' % cset
                raise ExclusiveAttributesPresent(error)

    @property
    def raw(self):
        """Display the attributes that the current object has and their values.

        :returns: A dictionary of attributes and their values
        """
        return self.__dict__


class ResourceBase(PathElement, ToDictMixin):
    """Base class for all BIG-IP® iControl REST API endpoints.

    The BIG-IP® is represented by an object that converts device-published
    uri's into Python objects. Each uri maps to a Python object. The
    mechanism for instantiating these objects is the __getattr__ Special
    Function in the LazyAttributeMixin. When a registered attribute is `dot`
    referenced, on the device object (e.g. ``bigip.ltm`` or simply ``bigip``),
    an appropriate object is instantiated and attributed to the referencing
    object:

    .. code-block:: python

       bigip.ltm = LTM(bigip)
       bigip.ltm.nats
       nat1 = bigip.ltm.nats.nat.create('Foo', 'Bar', '0.1.2.3', '1.2.3.4')

    This can be shortened to just the last line:

    .. code-block:: python

       nat1 = bigip.ltm.nats.nat.create('Foo', 'Bar', '0.1.2.3', '1.2.3.4')

    Critically this enforces a convention relating device published uris to
    API objects, in a hierarchy similar to the uri paths.  I.E. the uri
    corresponding to a ``Nats`` object is ``mgmt/tm/ltm/nat/``. If you
    query the BIG-IP's uri (e.g. print(bigip._meta_data['uri']) ), you'll see
    that it ends in:
    ``/mgmt/tm/``, if you query the ``ltm`` object's uri
    (e.g. print(bigip.ltm._meta_data['uri']) ) you'll see it ends in
    ``/mgmt/tm/ltm/``.

    In general the objects build a required `self._meta_data['uri']` attribute
    by:
    1. Inheriting this class.
    2. calling super(Subclass, self).__init__(container)
    3. self.uri = self.container_uri['uri'] + '/' + self.__class__.__name__

    The net result is a succinct mapping between uri's and objects,
    that represents objects in a hierarchical relationship similar to the
    device's uri path hierarchy.
    """
    def _modify(self, **patch):
        """Wrapped with modify, override in a subclass to customize."""

        requests_params, patch_uri, session, read_only = \
            self._prepare_put_or_patch(patch)
        self._check_for_boolean_pair_reduction(patch)
        read_only_mutations = []
        for attr in read_only:
            if attr in patch:
                read_only_mutations.append(attr)
        if read_only_mutations:
            msg = 'Attempted to mutate read-only attribute(s): %s' \
                % read_only_mutations
            raise AttemptedMutationOfReadOnly(msg)

        patch = self._prepare_request_json(patch)
        response = session.patch(patch_uri, json=patch, **requests_params)
        self._local_update(response.json())

    def modify(self, **patch):
        """Modify the configuration of the resource on device based on patch

        """
        self._modify(**patch)

    def _check_for_boolean_pair_reduction(self, kwargs):
        """Check if boolean pairs should be reduced in this resource."""

        if 'reduction_forcing_pairs' in self._meta_data:
            for key1, key2 in self._meta_data['reduction_forcing_pairs']:
                kwargs = self._reduce_boolean_pair(kwargs, key1, key2)
        return kwargs

    def _prepare_put_or_patch(self, kwargs):
        """Retrieve the appropriate request items for put or patch calls."""

        requests_params = self._handle_requests_params(kwargs)
        update_uri = self._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']
        read_only = self._meta_data.get('read_only_attributes', [])
        return requests_params, update_uri, session, read_only

    def _prepare_request_json(self, kwargs):
        """Prepare request args for sending to device as JSON."""

        # Check for python keywords in dict
        kwargs = self._check_for_python_keywords(kwargs)

        # Check for the key 'check' in kwargs
        if 'check' in kwargs:
            od = OrderedDict()
            od['check'] = kwargs['check']
            kwargs.pop('check')
            od.update(kwargs)
            return od
        return kwargs

    def _iter_list_for_dicts(self, check_list):
        """Iterate over list to find dicts and check for python keywords."""

        list_copy = copy.deepcopy(check_list)
        for index, elem in enumerate(check_list):
            if isinstance(elem, dict):
                list_copy[index] = self._check_for_python_keywords(elem)
            elif isinstance(elem, list):
                list_copy[index] = self._iter_list_for_dicts(elem)
            else:
                list_copy[index] = elem
        return list_copy

    def _check_for_python_keywords(self, kwargs):
        """When Python keywords seen, mutate to remove trailing underscore."""

        kwargs_copy = copy.deepcopy(kwargs)
        for key, val in iteritems(kwargs):
            if isinstance(val, dict):
                kwargs_copy[key] = self._check_for_python_keywords(val)
            elif isinstance(val, list):
                kwargs_copy[key] = self._iter_list_for_dicts(val)
            else:
                if key.endswith('_'):
                    strip_key = key.rstrip('_')
                    if keyword.iskeyword(strip_key):
                        kwargs_copy[strip_key] = val
                        kwargs_copy.pop(key)
        return kwargs_copy

    def _check_keys(self, rdict):
        """Call this from _local_update to validate response keys

        disallowed server-response json keys:
        1. The string-literal '_meta_data'
        2. strings that are not valid Python 2.7 identifiers
        3. strings beginning with '__'.

        :param rdict: from response.json()
        :raises: DeviceProvidesIncompatibleKey
        :returns: checked response rdict
        """
        if '_meta_data' in rdict:
            error_message = "Response contains key '_meta_data' which is "\
                "incompatible with this API!!\n Response json: %r" % rdict
            raise DeviceProvidesIncompatibleKey(error_message)
        for x in rdict:
            if not re.match(tokenize.Name, x):
                error_message = "Device provided %r which is disallowed"\
                    " because it's not a valid Python 2.7 identifier." % x
                raise DeviceProvidesIncompatibleKey(error_message)
            elif keyword.iskeyword(x):
                # If attribute is keyword, append underscore to attribute name
                rdict[x + '_'] = rdict[x]
                rdict.pop(x)
            elif x.startswith('__'):
                error_message = "Device provided %r which is disallowed"\
                    ", it mangles into a Python non-public attribute." % x
                raise DeviceProvidesIncompatibleKey(error_message)
        return rdict

    def _local_update(self, rdict):
        """Call this with a response dictionary to update instance attrs.

        If the response has only valid keys, stash meta_data, replace __dict__,
        and reassign meta_data.

        :param rdict: response attributes derived from server JSON
        """
        sanitized = self._check_keys(rdict)
        temp_meta = self._meta_data
        self.__dict__ = sanitized
        self._meta_data = temp_meta

    def _update(self, **kwargs):
        """wrapped with update, override that in a subclass to customize"""

        requests_params, update_uri, session, read_only = \
            self._prepare_put_or_patch(kwargs)

        read_only_mutations = []
        for attr in read_only:
            if attr in kwargs:
                read_only_mutations.append(attr)
        if read_only_mutations:
            msg = 'Attempted to mutate read-only attribute(s): %s' \
                  % read_only_mutations
            raise AttemptedMutationOfReadOnly(msg)

        # Get the current state of the object on BIG-IP® and check the
        # generation Use pop here because we don't want force in the data_dict
        force = self._check_force_arg(kwargs.pop('force', True))
        if not force:
            # generation has a known server-side error
            self._check_generation()

        kwargs = self._check_for_boolean_pair_reduction(kwargs)

        # Save the meta data so we can add it back into self after we
        # load the new object.
        temp_meta = self.__dict__.pop('_meta_data')

        # Need to remove any of the Collection objects from self.__dict__
        # because these are subCollections and _meta_data and
        # other non-BIG-IP® attrs are not removed from the subCollections
        # See issue #146 for details
        tmp = dict()
        for key, value in iteritems(self.__dict__):
            # In Python2 versions we were changing a dictionary in place,
            # but this cannot be done with an iterator as an error is raised.
            # So instead we create a temporary holder for the modified dict
            # and then re-assign it afterwards.
            if isinstance(value, Collection):
                pass
            else:
                tmp[key] = value
        self.__dict__ = tmp
        data_dict = self.to_dict()

        # Remove any read-only attributes from our data_dict before we update
        # the data dict with the attributes.  If they pass in read-only attrs
        # in the method call we are going to let BIG-IP® let them know about it
        # when it fails
        for attr in read_only:
            data_dict.pop(attr, '')

        data_dict.update(kwargs)
        data_dict = self._prepare_request_json(data_dict)

        # Handles ConnectionAborted errors
        #
        # @see https://github.com/F5Networks/f5-ansible/issues/317
        # @see https://github.com/requests/requests/issues/2364
        for _ in range(0, 30):
            try:
                response = session.put(update_uri, json=data_dict, **requests_params)
                self._meta_data = temp_meta
                self._local_update(response.json())
                break
            except iControlUnexpectedHTTPError:
                response = session.get(update_uri, **requests_params)
                self._meta_data = temp_meta
                self._local_update(response.json())
                raise
            except ConnectionError as ex:
                if 'Connection aborted' in str(ex):
                    time.sleep(1)
                    continue
                else:
                    raise

    def update(self, **kwargs):
        """Update the configuration of the resource on the BIG-IP®.

        This method uses HTTP PUT alter the resource state on the BIG-IP®.

        The attributes of the instance will be packaged as a dictionary.  That
        dictionary will be updated with kwargs.  It is then submitted as JSON
        to the device.

        Various edge cases are handled:
        * read-only attributes that are unchangeable are removed

        Args:
            kwargs (dict): Arbitrary number of keyword arguments.
                Keys and associated values to alter on the device.

                If kwargs has a ``requests_params`` key the corresponding dict will
                be passed to the underlying ``requests.session.put`` method where it will
                be handled according to that API.
        """
        # Need to implement checking for valid params here.
        self._update(**kwargs)

    def _refresh(self, **kwargs):
        """wrapped by `refresh` override that in a subclass to customize"""
        requests_params = self._handle_requests_params(kwargs)
        refresh_session = self._meta_data['bigip']._meta_data['icr_session']

        if self._meta_data['uri'].endswith('/stats/'):
            # Slicing off the trailing slash here for Stats enpoints because
            # iWorkflow doesn't consider those `stats` URLs valid if they
            # include the trailing slash.
            #
            # Other than that, functionality does not change
            uri = self._meta_data['uri'][0:-1]
        else:
            uri = self._meta_data['uri']

        response = refresh_session.get(uri, **requests_params)
        self._local_update(response.json())

    def refresh(self, **kwargs):
        """Use this to make the device resource be represented by self.

        This method makes an HTTP GET query against the device service.
        This method is run for its side-effects on self.
        If successful the instance attribute __dict__ is replaced
        with the dict representing the device state.  To figure out what that
        state is, run a subsequest query of the object like this:
        As with all CURDLE methods use a "requests_params" dict to pass
        parameters to requests.session.HTTPMETHOD. See test_requests_params.py
        for an example.
        >>> resource_obj.refresh()
        >>> print(resource_obj.raw)
        """
        self._refresh(**kwargs)

    def load(self, **kwargs):
        error_message = "Only Resources support 'load'."
        raise InvalidResource(error_message)

    def create(self, **kwargs):
        """Implement this by overriding it in a subclass of `Resource`

        Raises:
            InvalidResource: If method is used.
        """
        error_message = "Only Resources support 'create'."
        raise InvalidResource(error_message)

    def delete(self, **kwargs):
        """Implement this by overriding it in a subclass of `Resource`

        Raises:
            InvalidResource: If method is used.
        """
        error_message = "Only Resources support 'delete'."
        raise InvalidResource(error_message)

    def _stamp_out_core(self):
        container = self._meta_data['container']
        core = self.__class__.__new__(self.__class__)
        core.__init__(container)
        return core

    def _produce_instance(self, response):
        '''Generate a new self, which is an instance of the self.'''
        new_instance = self._stamp_out_core()
        # Post-process the response
        new_instance._local_update(response.json())

        if new_instance.kind != new_instance._meta_data['required_json_kind'] \
           and new_instance.kind != "tm:transaction:commandsstate":
            error_message = "For instances of type '%r' the corresponding"\
                " kind must be '%r' but creation returned JSON with kind: %r"\
                % (new_instance.__class__.__name__,
                   new_instance._meta_data['required_json_kind'],
                   new_instance.kind)
            raise KindTypeMismatch(error_message)

        # Update the object to have the correct functional uri.
        new_instance._activate_URI(new_instance.selfLink)
        return new_instance

    def _reduce_boolean_pair(self, config_dict, key1, key2):
        """Ensure only one key with a boolean value is present in dict.

        :param config_dict: dict -- dictionary of config or kwargs
        :param key1: string -- first key name
        :param key2: string -- second key name
        :raises: BooleansToReduceHaveSameValue
        """

        if key1 in config_dict and key2 in config_dict \
                and config_dict[key1] == config_dict[key2]:
            msg = 'Boolean pair, %s and %s, have same value: %s. If both ' \
                'are given to this method, they cannot be the same, as this ' \
                'method cannot decide which one should be True.' \
                % (key1, key2, config_dict[key1])
            raise BooleansToReduceHaveSameValue(msg)
        elif key1 in config_dict and not config_dict[key1]:
            config_dict[key2] = True
            config_dict.pop(key1)
        elif key2 in config_dict and not config_dict[key2]:
            config_dict[key1] = True
            config_dict.pop(key2)
        return config_dict

    @property
    def attrs(self):
        no_meta_dict = {k: v for k, v in iteritems(self.__dict__)
                        if k != '_meta_data'}
        return no_meta_dict


class OrganizingCollection(ResourceBase):
    """Base class for objects that collect resources under them.

    ``OrganizingCollection`` objects fulfill the following functions,

    * represent a uri path fragment immediately 'below' ``/mgmt/tm``
    * provide a list of dictionaries that contain uri's to other
      resources on the device.
    """

    def get_collection(self, **kwargs):
        """Call to obtain a list of the reference dicts in the instance `items`

        Returns:
            list: List of self.items
        """
        self.refresh(**kwargs)
        return self.items


class Collection(ResourceBase):
    """Base class for objects that collect a list of ``Resources``

    The Collection Resource is responsible for providing a list of Python
    objects, where each object represents a unique URI, the URI contains the
    URI of the Collection at the front of its path, and the 'kind' of the
    URI-associated-JSON has been registered with the attribute registry of the
    Collection subclass.

    .. note::

        Any subclass of this base class must have ``s`` at the end of its name
        unless it ends in ``s`` then it must have ``_s``.

    """

    def get_collection(self, **kwargs):
        """Get an iterator of Python ``Resource`` objects that represent URIs.

        The returned objects are Pythonic `Resource`s that map to the most
        recently `refreshed` state of uris-resources published by the device.
        In order to instantiate the correct types, the concrete subclass must
        populate its registry with acceptable types, based on the `kind` field
        returned by the REST server.

        .. note::
            This method implies a single REST transaction with the
            Collection subclass URI.

        :raises: UnregisteredKind
        :returns: list of reference dicts and Python ``Resource`` objects
        """
        list_of_contents = []
        self.refresh(**kwargs)
        if 'items' in self.__dict__:
            for item in self.items:
                # It's possible to have non-"kind" JSON returned. We just
                # append the corresponding dict. PostProcessing is the caller's
                # responsibility.
                if 'kind' not in item:
                    list_of_contents.append(item)
                    continue
                kind = item['kind']
                if kind in self._meta_data['attribute_registry']:
                    # If it has a kind, it must be registered.
                    instance = self._meta_data['attribute_registry'][kind](self)
                    instance._local_update(item)
                    instance._activate_URI(instance.selfLink)
                    list_of_contents.append(instance)
                else:
                    error_message = '%r is not registered!' % kind
                    raise UnregisteredKind(error_message)
        return list_of_contents


class Resource(ResourceBase):
    """Base class to represent a Configurable Resource on the device.

    .. warning::

       Objects instantiated from subclasses of Resource do NOT contain a URI
       (self._meta_data['uri']) at instantiation!

    Resource objects provide the interface for the Creation of new services on
    the device. Once a new service has been created, (via ``self.create`` or
    ``self.load``), the instance constructs its URI and stores it as
    ``self._meta_data['uri']``.

    It is an error to attempt to call
    :meth:`~f5.bigip.resource.Resource.create` or
    :meth:`~f5.bigip.resource.Resource.load` on an instance more than once.
    ``self._meta_data['uri']`` MUST not be changed after creation or load.

    .. note::

       creation query args, and creation hash fragments are stored as
       separate _meta_data values.

    By "Configurable" we mean that submitting JSON via the PUT method to the
    URI managed by subclasses of Resource, changes the state of the
    corresponding service on the device.

    It also means that the URI supports `DELETE`.
    """
    def __init__(self, container):
        """Call to create a client side object to represent a service URI.

        Call _create or _load for a Resource resource to have a
        self._meta_data['uri']!
        """
        super(Resource, self).__init__(container)
        self._meta_data.pop('uri')
        # Creation fails without these.
        self._meta_data['required_creation_parameters'] = {'name', }
        # Refresh fails without these.
        self._meta_data['required_load_parameters'] = {'name', }
        # You can't set these attributes, only 'read' them.
        self._meta_data['read_only_attributes'] = []
        self._meta_data['reduction_forcing_pairs'] = \
            [
                ('enabled', 'disabled'),
                ('online', 'offline'),
                ('vlansEnabled', 'vlansDisabled')
            ]

    def _activate_URI(self, selfLinkuri):
        """Call this with a selfLink, after it's returned in _create or _load.

        Each instance is tightly bound to a particular service URI.  When that
        service is created by this library, or loaded from the device, the URI
        is set to self._meta_data['uri'].   This operation can only occur once,
        any subsequent attempt to manipulate self._meta_data['uri'] is
        probably a mistake.

        self.selfLink references a value that is returned as a JSON value from
        the device.  This value contains "localhost" as the domain or the uri.
        "localhost" is only conceivably useful if the client library is run on
        the device itself, so it is replaced with the domain this API used to
        communicate with the device.

        self.selfLink correctly contains a complete uri, that is only _now_
        (post create or load) available to self.

        Now that the complete URI is available to self, it is now possible to
        reference subcollections, as attributes of self!
        e.g. a resource with a uri path like:
        "/mgmt/tm/ltm/pool/~Common~pool_collection1/members"
        The mechanism used to enable this change is to set
        the `allowed_lazy_attributes` _meta_data key to hold values of the
        `attribute_registry` _meta_data key.

        Finally we stash the corrected `uri`, returned hash_fragment, query
        args, and of course allowed_lazy_attributes in _meta_data.

        :param selfLinkuri: the server provided selfLink (contains localhost)
        :raises: URICreationCollision
        """

        # netloc local alias
        uri = urlparse.urlsplit(str(self._meta_data['bigip']._meta_data['uri']))

        # attrs local alias
        attribute_reg = self._meta_data.get('attribute_registry', {})
        attrs = list(itervalues(attribute_reg))
        attrs = self._assign_stats(attrs)

        (scheme, domain, path, qarg, frag) = urlparse.urlsplit(selfLinkuri)
        path_uri = urlparse.urlunsplit((scheme, uri.netloc, path, '', ''))
        if not path_uri.endswith('/'):
            path_uri = path_uri + '/'
        qargs = urlparse.parse_qs(qarg)
        self._meta_data.update({'uri': path_uri,
                                'creation_uri_qargs': qargs,
                                'creation_uri_frag': frag,
                                'allowed_lazy_attributes': attrs})

    def _assign_stats(self, attrs):
        if self._meta_data['object_has_stats']:
            attrs.append(Stats)
        return attrs

    def _check_create_parameters(self, **kwargs):
        """Params given to create should satisfy required params.

        :params: kwargs
        :raises: MissingRequiredCreateParameter
        """
        rset = self._meta_data['required_creation_parameters']
        check = _missing_required_parameters(rset, **kwargs)
        if check:
            error_message = 'Missing required params: %s' % check
            raise MissingRequiredCreationParameter(error_message)

    def _minimum_one_is_missing(self, **kwargs):
        """Helper function to do operation on sets

        Verify if at least one of the elements
        is present in **kwargs. If no items of rqset
        are contained in **kwargs  the function
        raises exception.

        This check will only trigger if rqset is not empty.

        Raises:
             MissingRequiredCreationParameter
        """
        rqset = self._meta_data['minimum_additional_parameters']
        if rqset:
            kwarg_set = set(iterkeys(kwargs))
            if kwarg_set.isdisjoint(rqset):
                args = sorted(rqset)
                error_message = 'This resource requires at least one of the ' \
                                'mandatory additional ' \
                                'parameters to be provided: %s' % ', '.join(args)
                raise MissingRequiredCreationParameter(error_message)

    def _create(self, **kwargs):
        """wrapped by `create` override that in subclasses to customize"""
        if 'uri' in self._meta_data:
            error = "There was an attempt to assign a new uri to this "\
                    "resource, the _meta_data['uri'] is %s and it should"\
                    " not be changed." % (self._meta_data['uri'])
            raise URICreationCollision(error)
        self._check_exclusive_parameters(**kwargs)
        requests_params = self._handle_requests_params(kwargs)
        self._minimum_one_is_missing(**kwargs)
        self._check_create_parameters(**kwargs)
        kwargs = self._check_for_python_keywords(kwargs)

        # Reduce boolean pairs as specified by the meta_data entry below
        for key1, key2 in self._meta_data['reduction_forcing_pairs']:
            kwargs = self._reduce_boolean_pair(kwargs, key1, key2)

        # Make convenience variable with short names for this method.
        _create_uri = self._meta_data['container']._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']

        kwargs = self._prepare_request_json(kwargs)

        # Invoke the REST operation on the device.
        response = session.post(_create_uri, json=kwargs, **requests_params)

        # Make new instance of self
        result = self._produce_instance(response)
        return result

    def create(self, **kwargs):
        r"""Create the resource on the BIG-IP®.

        Uses HTTP POST to the `collection` URI to create a resource associated
        with a new unique URI on the device.

        ..
            Subclasses can customize  this functionality by defining a `create`
            method that wraps and calls this method with appropriate arguments.

            Note this is the one of two fundamental Resource operations that
            returns a different uri (in the returned object) than the uri the
            operation was called on.  The returned uri can be accessed as
            Object.selfLink, the actual uri used by REST operations on the
            object is Object._meta_data['uri'].  The _meta_data['uri'] is the
            same as Object.selfLink with the substring 'localhost' replaced
            with the value of
            Object._meta_data['bigip']._meta_data['hostname'], and without
            query args, or hash fragments.

        Args:
            \*\*kwargs (dict): Arbitrary number of keyword arguments.
                All the key-values needed to create the resource.

                If kwargs has a ``requests_params`` key the corresponding dict will
                be passed to the underlying ``requests.session.post`` method where it
                will be handled according to that API.

        Returns:
            Resource: A python object that represents the object's configuration
                and state on the BIG-IP®.
        """
        return self._create(**kwargs)

    def _check_load_parameters(self, **kwargs):
        """Params given to load should at least satisfy required params.

        :params: kwargs
        :raises: MissingRequiredReadParameter
        """
        rset = self._meta_data['required_load_parameters']
        check = _missing_required_parameters(rset, **kwargs)
        if check:
            check.sort()
            error_message = 'Missing required params: %s' % check
            raise MissingRequiredReadParameter(error_message)

    def _load(self, **kwargs):
        """wrapped with load, override that in a subclass to customize"""
        if 'uri' in self._meta_data:
            error = "There was an attempt to assign a new uri to this "\
                    "resource, the _meta_data['uri'] is %s and it should"\
                    " not be changed." % (self._meta_data['uri'])
            raise URICreationCollision(error)
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        refresh_session = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']
        kwargs.update(requests_params)
        for key1, key2 in self._meta_data['reduction_forcing_pairs']:
            kwargs = self._reduce_boolean_pair(kwargs, key1, key2)
        kwargs = self._check_for_python_keywords(kwargs)
        response = refresh_session.get(base_uri, **kwargs)
        # Make new instance of self
        return self._produce_instance(response)

    def load(self, **kwargs):
        r"""Load an already configured service into this instance.

        This method uses HTTP GET to obtain a resource from the BIG-IP®.

        The URI of the target service is constructed from the instance's
        container and ``**kwargs``.

        ``kwargs`` typically requires the keys ``name`` and ``partition``.
        this may, or may not, be true for a specific service.

        Args:
            \*\*kwargs (dict): Arbitrary number of keyword arguments.
                If kwargs has a ``requests_params`` key the corresponding dict will
                be passed to the underlying ``requests.session.get`` method where it will
                be handled according to that API.

                Use the method above to pass query args

        Returns:
            Resource: A Resource Instance with a populated ``_meta_data['uri']``
        """
        return self._load(**kwargs)

    def _delete(self, **kwargs):
        """wrapped with delete, override that in a subclass to customize """
        requests_params = self._handle_requests_params(kwargs)

        delete_uri = self._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']

        # Check the generation for match before delete
        force = self._check_force_arg(kwargs.pop('force', True))
        if not force:
            self._check_generation()

        response = session.delete(delete_uri, **requests_params)
        if response.status_code == 200:
            self.__dict__ = {'deleted': True}

    def delete(self, **kwargs):
        r"""Delete the resource on the BIG-IP®.

        Uses HTTP DELETE to delete the resource on the BIG-IP®.

        After this method is called, and status_code 200 response is received
        ``instance.__dict__`` is replace with ``{'deleted': True}``

        Args:
            \*\*kwargs (dict): Arbitrary number of keyword arguments.
                The only current use is to pass kwargs to the requests API.

                If kwargs has a ``requests_params`` key the corresponding dict will
                be passed to the underlying ``requests.session.delete`` method where it
                will be handled according to that API.

                Use the method above to pass query args.
        """
        # Need to implement checking for ? here.
        self._delete(**kwargs)
        # Need to implement correct teardown here.

    def exists(self, **kwargs):
        r"""Check for the existence of the named object on the BIG-IP

        Sends an HTTP GET to the URI of the named object and if it fails with
        a :exc:~requests.HTTPError` exception it checks the exception for
        status code of 404 and returns :obj:`False` in that case.

        If the GET is successful it returns :obj:`True`.

        For any other errors are raised as-is.

        Args:
            \*\*kwargs (dict): Arbitrary number of keyword arguments.
                If kwargs has a ``requests_params`` key the corresponding dict will
                be passed to the underlying ``requests.session.get`` method where it will
                be handled according to that API.

                Use the method above to pass query args.

        Returns:
             bool: True is the object exists. False otherwise.

        Raises:
             requests.HTTPError: Any HTTP error that was not status code 404.
        """
        return self._exists(**kwargs)

    def _exists(self, **kwargs):
        """wrapped with exists, override that in a subclass to customize """
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        session = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']
        kwargs.update(requests_params)

        try:
            session.get(base_uri, **kwargs)
        except HTTPError as err:
            if err.response.status_code == 404:
                return False
            else:
                raise
        return True


class UnnamedResource(ResourceBase):
    """This makes a resource object work if there is no name.

    These objects do not support create or delete and are often found
    as Resources that are under an organizing collection.

    For example the ``mgmt/tm/sys/global-settings`` is one of these and has a kind of
    ``tm:sys:global-settings:global-settingsstate`` and the URI does not match the kind.
    """

    def create(self, **kwargs):
        """Create is not supported for unnamed resources

        Raises:
            UnsupportedMethod: If the method is used
        """
        raise UnsupportedMethod(
            "%s does not support the create method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        """Delete is not supported for unnamed resources

        Raises:
            UnsupportedMethod: If the method is used
        """
        raise UnsupportedMethod(
            "%s does not support the delete method" % self.__class__.__name__
        )

    def load(self, **kwargs):
        newinst = self._stamp_out_core()
        newinst._refresh(**kwargs)
        return newinst


class Stats(UnnamedResource):
    """For stats resources."""

    def modify(self, **kwargs):
        """Modify is not supported for stats resources

        Raises:
            UnsupportedMethod: If the method is used
        """
        raise UnsupportedMethod(
            "%s does not support the modify method" % self.__class__.__name__
        )

    def load(self, **kwargs):
        # TODO(pjbreaux) add try-except and custom exception here.
        return super(Stats, self).load(**kwargs)


class AsmResource(Resource):
    """ASM Resource class represents a configurable ASM endpoint on the device.

    ASM resources are unique in BIG-IP® in the sense that their direct URI
    endpoints are hash based IDs of the resources.

    The IDs are generated by BIG-IP® when the objects are created.

    Moreover, the ASM resources do not have 'generation' property,
    therefore some of the other methods needed to be adjusted to accommodate
    that.
    """

    def __init__(self, container):
        """Call to create a client side object to represent a service URI.

        Call _create or _load for a Resource to have a self._meta_data['uri']!
        """
        super(AsmResource, self).__init__(container)
        # Asm endpoints require object 'id' which is a hash created by BIGIP
        # when object is created.
        self._meta_data['required_load_parameters'] = {'id', }
        # No ASM endpoint supports Stats
        self._meta_data['object_has_stats'] = False

    def _load(self, **kwargs):
        """wrapped with load, override that in a subclass to customize"""
        if 'uri' in self._meta_data:
            error = "There was an attempt to assign a new uri to this "\
                    "resource, the _meta_data['uri'] is %s and it should"\
                    " not be changed." % (self._meta_data['uri'])
            raise URICreationCollision(error)
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        refresh_session = self._meta_data['bigip']._meta_data['icr_session']
        uri = self._meta_data['container']._meta_data['uri']
        endpoint = kwargs.pop('id', '')
        # Popping name kwarg as it will cause the uri to be invalid. We only
        # require id parameter
        kwargs.pop('name', '')
        base_uri = uri + endpoint + '/'
        kwargs.update(requests_params)
        for key1, key2 in self._meta_data['reduction_forcing_pairs']:
            kwargs = self._reduce_boolean_pair(kwargs, key1, key2)
        response = refresh_session.get(base_uri, **kwargs)
        # Make new instance of self
        return self._produce_instance(response)

    def load(self, **kwargs):
        r"""Load an already configured service into this instance.

        This method uses HTTP GET to obtain a resource from the BIG-IP®.

        The URI of the target service is constructed from the instance's
        container and ``\*\*kwargs``.

        kwargs typically for ASM requires "id" in majority of cases,
        as object links in ASM are using hash(id) instead of names,
        this may, or may not, be true for a specific service.

        Args:
            \*\*kwargs (dict): Arbitrary number of keyword arguments.
                If kwargs has a ``requests_params`` key the corresponding dict will
                be passed to the underlying ``requests.session.get`` method where it will
                be handled according to that API.

        Returns:
            Resource: ``Resource`` instance with a populated _meta_data['uri']
        """
        return self._load(**kwargs)

    def _delete(self, **kwargs):
        """Wrapped with delete, override that in a subclass to customize """
        requests_params = self._handle_requests_params(kwargs)

        delete_uri = self._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']
        response = session.delete(delete_uri, **requests_params)
        if response.status_code == 200 or 201:
            self.__dict__ = {'deleted': True}

    def delete(self, **kwargs):
        """Delete the ASM resource on the BIG-IP®.

        Uses HTTP DELETE to delete the ASM resource on the BIG-IP®.

        After this method is called, and status_code 200 or 201 response is
        received ``instance.__dict__`` is replace with ``{'deleted': True}``

        Args:
            kwargs (dict): Arbitrary number of keyword arguments.
                If kwargs has a ``requests_params`` key the corresponding dict will
                be passed to the underlying ``requests.session.delete`` method where it
                will be handled according to that API.
        """
        # Need to implement checking for ? here.
        self._delete(**kwargs)
        # Need to implement correct teardown here.

    def exists(self, **kwargs):
        r"""Check for the existence of the ASM object on the BIG-IP

        Sends an HTTP GET to the URI of the ASM object and if it fails with
        a :exc:~requests.HTTPError` exception it checks the exception for
        status code of 404 and returns :obj:`False` in that case.

        If the GET is successful it returns :obj:`True`.

        For any other errors are raised as-is.

        Args:
            \*\*kwargs (dict): Arbitrary number of keyword arguments.
                Keyword arguments required to get objects

                If kwargs has a ``requests_param`` key the corresponding dict will
                be passed to the underlying ``requests.session.get`` method where it will
                be handled according to that API.

        Returns:
            bool: True is the object exists: False otherwise.

        Raises:
            requests.HTTPError: Any HTTP error that was not status code 404.
        """
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        session = self._meta_data['bigip']._meta_data['icr_session']
        uri = self._meta_data['container']._meta_data['uri']
        endpoint = kwargs.pop('id', '')
        # Popping name kwarg as it will cause the uri to be invalid
        kwargs.pop('name', '')
        base_uri = uri + endpoint + '/'
        kwargs.update(requests_params)
        try:
            session.get(base_uri, **kwargs)
        except HTTPError as err:
            if err.response.status_code == 404:
                return False
            else:
                raise
        return True

    def update(self, **kwargs):
        """Update is not supported for ASM Resources

        Raises:
            UnsupportedOperation: If method is used.
        """
        raise UnsupportedOperation(
            "%s does not support the update method" % self.__class__.__name__
        )


class AsmTaskResource(AsmResource):
    """ASM Task Resource class represents an ASM Tasks endpoint on the

    device.

    Tasks resources do not support create() method in the strict sense,
    as they require an HTTP POST with an empty json{} to prompt BIGIP to
    create them, therefore a new method fetch() was created.

    """

    def __init__(self, container):
        """Call to create a client side object to represent a service URI.

        Call  _fetch for a Resource to have a self._meta_data['uri']!
        """
        super(AsmTaskResource, self).__init__(container)

    def _fetch(self):
        """wrapped by `fetch` override that in subclasses to customize"""
        if 'uri' in self._meta_data:
            error = "There was an attempt to assign a new uri to this "\
                    "resource, the _meta_data['uri'] is %s and it should"\
                    " not be changed." % (self._meta_data['uri'])
            raise URICreationCollision(error)
        # Make convenience variable with short names for this method.
        _create_uri = self._meta_data['container']._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']
        # Invoke the REST operation on the device.
        response = session.post(_create_uri, json={})
        # Make new instance of self
        return self._produce_instance(response)

    def fetch(self):
        """Fetch the ASM resource on the BIG-IP®.

        This is a heavily modified version of create, that does not allow
        any arguments when executing. It uses an emtpy json{} HTTP POST to
        prompt the BIG-IP® to create the object, mainly used by 'Tasks'
        endpoint.
        """
        return self._fetch()

    def create(self, **kwargs):
        """Create is not supported for Task ASM resources

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the create method" % self.__class__.__name__
        )

    def modify(self, **kwargs):
        """Modify is not supported for Task ASM resources

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )


class TaskResource(Resource):
    def __init__(self, container):
        """Call to create a client side object to represent a service URI.

        Call _create or _load for a Resource to have a self._meta_data['uri']!
        """
        super(TaskResource, self).__init__(container)
        # Asm endpoints require object 'id' which is a hash created by BIGIP
        # when object is created.
        self._meta_data['required_load_parameters'] = {'id', }
        # No ASM endpoint supports Stats
        self._meta_data['object_has_stats'] = False

    def _load(self, **kwargs):
        """wrapped with load, override that in a subclass to customize"""
        if 'uri' in self._meta_data:
            error = "There was an attempt to assign a new uri to this " \
                    "resource, the _meta_data['uri'] is %s and it should" \
                    " not be changed." % (self._meta_data['uri'])
            raise URICreationCollision(error)
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        refresh_session = self._meta_data['bigip']._meta_data['icr_session']
        uri = self._meta_data['container']._meta_data['uri']
        endpoint = kwargs.pop('id', '')
        # Popping name kwarg as it will cause the uri to be invalid. We only
        # require id parameter
        kwargs.pop('name', '')
        base_uri = uri + endpoint + '/'
        kwargs.update(requests_params)
        for key1, key2 in self._meta_data['reduction_forcing_pairs']:
            kwargs = self._reduce_boolean_pair(kwargs, key1, key2)
        response = refresh_session.get(base_uri, **kwargs)
        # Make new instance of self
        return self._produce_instance(response)

    def load(self, **kwargs):
        """Load an already configured service into this instance.

        This method uses HTTP GET to obtain a resource from the BIG-IP®.

        ..
            The URI of the target service is constructed from the instance's
            container and **kwargs.
            kwargs typically for ASM requires "id" in majority of cases,
            as object links in ASM are using hash(id) instead of names,
            this may, or may not, be true for a specific service.

        :param kwargs: typically contains "id"
        NOTE: If kwargs has a 'requests_params' key the corresponding dict will
        be passed to the underlying requests.session.get method where it will
        be handled according to that API. THIS IS HOW TO PASS QUERY-ARGS!
        :returns: a Resource Instance (with a populated _meta_data['uri'])
        """
        return self._load(**kwargs)

    def _delete(self, **kwargs):
        """wrapped with delete, override that in a subclass to customize """
        requests_params = self._handle_requests_params(kwargs)

        delete_uri = self._meta_data['uri']
        session = self._meta_data['bigip']._meta_data['icr_session']
        response = session.delete(delete_uri, **requests_params)
        if response.status_code == 200 or 201:
            self.__dict__ = {'deleted': True}

    def delete(self, **kwargs):
        """Delete the Task resource on the BIG-IP®.

        Uses HTTP DELETE to delete the Task resource on the BIG-IP®.

        After this method is called, and status_code 200 or 201 response is
        received ``instance.__dict__`` is replace with ``{'deleted': True}``

        :param kwargs: The only current use is to pass kwargs to the requests
        API. If kwargs has a 'requests_params' key the corresponding dict will
        be passed to the underlying requests.session.delete method where it
        will be handled according to that API. THIS IS HOW TO PASS QUERY-ARGS!
        """
        # Need to implement checking for ? here.
        self._delete(**kwargs)
        # Need to implement correct teardown here.

    def exists(self, **kwargs):
        r"""Check for the existence of the Task object on the BIG-IP

        Sends an HTTP GET to the URI of the ASM object and if it fails with
        a :exc:~requests.HTTPError` exception it checks the exception for
        status code of 404 and returns :obj:`False` in that case.

        If the GET is successful it returns :obj:`True`.

        For any other errors are raised as-is.

        Args:
            \*\*kwargs (dict): Arbitrary number of keyword arguments.
                Keyword arguments required to get objects.

                If kwargs has a ``requests_params`` key the corresponding dict will
                be passed to the underlying ``requests.session.get`` method where it will
                be handled according to that API.

        Returns:
            bool: True if the objects exists. False otherwise.

        Raises:
            requests.HTTPError: Any HTTP error that was not status code 404.
        """
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        session = self._meta_data['bigip']._meta_data['icr_session']
        uri = self._meta_data['container']._meta_data['uri']
        endpoint = kwargs.pop('id', '')
        # Popping name kwarg as it will cause the uri to be invalid
        kwargs.pop('name', '')
        base_uri = uri + endpoint + '/'
        kwargs.update(requests_params)
        try:
            session.get(base_uri, **kwargs)
        except HTTPError as err:
            if err.response.status_code == 404:
                return False
            else:
                raise
        return True

    def update(self, **kwargs):
        """Update is not supported for ASM Resources

        Raises:
            UnsupportedOperation: If method is used.
        """
        raise UnsupportedOperation(
            "%s does not support the update method" % self.__class__.__name__
        )



# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from f5.bigip import BigIP
from f5.bigip import ManagementRoot


def pytest_addoption(parser):
    parser.addoption("--bigip", action="store",
                     help="BIG-IP hostname or IP address")
    parser.addoption("--username", action="store", help="BIG-IP REST username",
                     default="admin")
    parser.addoption("--password", action="store", help="BIG-IP REST password",
                     default="admin")
    parser.addoption("--port", action="store", help="BIG-IP port",
                     default=443)
    parser.addoption("--token", action="store_true",
                     help="Token Authentication")
    parser.addoption("--peer", action="store",
                     help="Peer BIG-IP hostname or IP address", default='none')
    parser.addoption("--release", action="store",
                     help="TMOS version, in dotted format, eg. 12.0.0",
                     default='11.6.0')
    parser.addoption("--vcmp-host", action="store",
                     help="IP address of VCMP enabled host.")


@pytest.fixture(scope='session')
def opt_bigip(request):
    return request.config.getoption("--bigip")


@pytest.fixture(scope='session')
def opt_username(request):
    return request.config.getoption("--username")


@pytest.fixture(scope='session')
def opt_password(request):
    return request.config.getoption("--password")


@pytest.fixture(scope='session')
def opt_port(request):
    return request.config.getoption("--port")


@pytest.fixture(scope='session')
def opt_token(request):
    return request.config.getoption("--token")


@pytest.fixture(scope='session')
def opt_vcmp_host(request):
    return request.config.getoption("--vcmp-host")


@pytest.fixture(scope='session')
def opt_release(request):
    return request.config.getoption("--release")


@pytest.fixture
def opt_peer(request):
    return request.config.getoption("--peer")


@pytest.fixture(scope='session')
def mgmt_root(opt_bigip, opt_username, opt_password, opt_port, opt_token):
    '''bigip fixture'''
    try:
        from pytest import symbols
    except ImportError:
        m = ManagementRoot(opt_bigip, opt_username, opt_password,
                           port=opt_port, token=opt_token)
    else:
        if symbols is not None:
            m = ManagementRoot(symbols.bigip_mgmt_ip_public,
                               symbols.bigip_username,
                               symbols.bigip_password,
                               port=opt_port, token=opt_token)
        else:
            m = ManagementRoot(opt_bigip, opt_username, opt_password,
                               port=opt_port, token=opt_token)
    return m


@pytest.fixture
def bigip(mgmt_root):
    '''bigip fixture'''
    return mgmt_root.tm


@pytest.fixture(scope='module')
def vcmp_host(opt_vcmp_host, opt_username, opt_password, opt_port):
    '''vcmp fixture'''
    m = ManagementRoot(
        opt_vcmp_host, opt_username, opt_password, port=opt_port)
    return m


@pytest.fixture
def peer(opt_peer, opt_username, opt_password, scope="module"):
    '''peer bigip fixture'''
    p = BigIP(opt_peer, opt_username, opt_password)
    return p



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP Local Traffic Manager (LTM) Traffic Class module.

REST URI
    ``https://localhost/mgmt/tm/ltm/traffic-class``

GUI Path
    ``Local Traffic --> Traffic Class``

REST Kind
    ``tm:ltm:traffic-class:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Traffic_Class_s(Collection):
    """BIG-IP® LTM Traffic Class collection"""
    def __init__(self, ltm):
        super(Traffic_Class_s, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Traffic_Class]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:traffic-class:traffic-classstate': Traffic_Class}


class Traffic_Class(Resource):
    """BIG-IP® LTM Traffic Class Resource"""
    def __init__(self, traffic_class_s):
        super(Traffic_Class, self).__init__(traffic_class_s)
        self._meta_data['required_json_kind'] = \
            'tm:ltm:traffic-class:traffic-classstate'
        self._meta_data['required_creation_parameters'].update((
            'classification',))



#!/usr/bin/env python
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import print_function
import os
import sys

from scripts.configure import add_package


def store(script, pkg_type, pkg_name):
    working = os.path.dirname(script)
    add_package(pkg_type, pkg_name, working=working)


def usage():
    print("%s <pkg_type> <pkg_name>" % sys.argv[0])
    exit(-1)


def main():
    if not len(sys.argv) == 3:
        usage()
    store(*sys.argv)


if __name__ == '__main__':
    main()
    exit(0)



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Local Traffic Manager™ (LTM®) pool module.

REST URI
    ``http://localhost/mgmt/tm/ltm/pool``

GUI Path
    ``Local Traffic --> Pools``

REST Kind
    ``tm:ltm:pools:*``
"""

from requests.exceptions import HTTPError

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import MemberStateModifyUnsupported


class Pools(Collection):
    """BIG-IP® LTM pool collection"""
    def __init__(self, ltm):
        super(Pools, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Pool]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:pool:poolstate': Pool}


class Pool(Resource):
    """BIG-IP® LTM pool resource"""
    def __init__(self, pool_s):
        super(Pool, self).__init__(pool_s)
        self._meta_data['required_json_kind'] = 'tm:ltm:pool:poolstate'
        self._meta_data['attribute_registry'] = {
            'tm:ltm:pool:memberscollectionstate': Members_s
        }

    @staticmethod
    def _format_monitor_parameter(param):
        """This is a workaround for a known issue ID645289, which affects

        all versions of TMOS at this time.
        """
        if '{' in param and '}':
            tmp = param.strip('}').split('{')
            monitor = ''.join(tmp).rstrip()
            return monitor
        else:
            return param

    def create(self, **kwargs):
        """Custom create method to implement monitor parameter formatting."""
        if 'monitor' in kwargs:
            value = self._format_monitor_parameter(kwargs['monitor'])
            kwargs['monitor'] = value
        return super(Pool, self)._create(**kwargs)

    def update(self, **kwargs):
        """Custom update method to implement monitor parameter formatting."""
        if 'monitor' in kwargs:
            value = self._format_monitor_parameter(kwargs['monitor'])
            kwargs['monitor'] = value
        elif 'monitor' in self.__dict__:
            value = self._format_monitor_parameter(self.__dict__['monitor'])
            self.__dict__['monitor'] = value
        return super(Pool, self)._update(**kwargs)

    def modify(self, **patch):
        """Custom modify method to implement monitor parameter formatting."""
        if 'monitor' in patch:
            value = self._format_monitor_parameter(patch['monitor'])
            patch['monitor'] = value
        return super(Pool, self)._modify(**patch)


class Members_s(Collection):
    """BIG-IP® LTM pool members sub-collection"""
    def __init__(self, pool):
        super(Members_s, self).__init__(pool)
        self._meta_data['allowed_lazy_attributes'] = [Members]
        self._meta_data['required_json_kind'] =\
            'tm:ltm:pool:members:memberscollectionstate'
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:pool:members:membersstate': Members}


class Members(Resource):
    """BIG-IP® LTM pool members sub-collection resource"""
    def __init__(self, members_s):
        super(Members, self).__init__(members_s)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:pool:members:membersstate'
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['read_only_attributes'].append('ephemeral')
        self._meta_data['read_only_attributes'].append('address')

    def _check_member_parameters(self, **kwargs):
        """See discussion in issue #985."""
        if 'fqdn' in kwargs:
            kwargs['fqdn'].pop('autopopulate', '')
            kwargs['fqdn'].pop('addressFamily', '')
        if 'fqdn' in self.__dict__:
            self.__dict__['fqdn'].pop('autopopulate', '')
            self.__dict__['fqdn'].pop('addressFamily', '')
        if 'state' in kwargs:
            if kwargs['state'] != 'user-up' and kwargs['state'] != \
                    'user-down':
                kwargs.pop('state')
        if 'state' in self.__dict__:
            if self.__dict__['state'] != 'user-up' and self.__dict__['state'] \
                    != 'user-down':
                self.__dict__.pop('state')
        if 'session' in kwargs:
            if kwargs['session'] != 'user-enabled' and kwargs['session'] != \
                    'user-disabled':
                kwargs.pop('session')
        if 'session' in self.__dict__:
            if self.__dict__['session'] != 'user-enabled' and \
                    self.__dict__['session'] != 'user-disabled':
                self.__dict__.pop('session')
        # Until we implement sanity checks for __dict__ this needs to stay here
        self.__dict__.pop('ephemeral', '')
        self.__dict__.pop('address', '')
        return kwargs

    def update(self, **kwargs):
        """Call this to change the configuration of the service on the device.

        This method uses HTTP PUT to alter the service state on the device.

        The attributes of the instance will be packaged as a dictionary.  That
        dictionary will be updated with kwargs.  It is then submitted as JSON
        to the device.  Various edge cases are handled:

        * read-only attributes that are unchangeable are removed
        * If ``fqdn`` is in the kwargs or set as an attribute, removes the
          ``autopopulate`` and ``addressFamily`` keys from it.

        :param kwargs: keys and associated values to alter on the device
        """
        checked = self._check_member_parameters(**kwargs)
        return super(Members, self)._update(**checked)

    def exists(self, **kwargs):
        """Check for the existence of the named object on the BigIP

        Sends an HTTP GET to the URI of the named object and if it fails with
        a :exc:~requests.HTTPError` exception it checks the exception for
        status code of 404 and returns :obj:`False` in that case.

        If the GET is successful it must then check the contents of the json
        contained in the response, this is because the "pool/... /members"
        resource provided by the server returns a status code of 200 for
        queries that do not correspond to an existing configuration.  Therefore
        this method checks for the presence of the "address" key in the
        response JSON...  of course, this means that exists depends on an
        unexpected idiosyncrancy of the server, and might break with version
        updates, edge cases, or other unpredictable changes.

        :param kwargs: Keyword arguments required to get objects, "partition"
        and "name" are required

        NOTE: If kwargs has a 'requests_params' key the corresponding dict will
        be passed to the underlying requests.session.get method where it will
        be handled according to that API. THIS IS HOW TO PASS QUERY-ARGS!
        :returns: bool -- The objects exists on BigIP or not.
        :raises: :exc:`requests.HTTPError`, Any HTTP error that was not status
                 code 404.
        """
        requests_params = self._handle_requests_params(kwargs)
        self._check_load_parameters(**kwargs)
        kwargs['uri_as_parts'] = True
        session = self._meta_data['bigip']._meta_data['icr_session']
        base_uri = self._meta_data['container']._meta_data['uri']
        kwargs.update(requests_params)
        try:
            response = session.get(base_uri, **kwargs)
        except HTTPError as err:
            if err.response.status_code == 404:
                return False
            else:
                raise
        rdict = response.json()
        if "address" not in rdict:
            # We can add 'or' conditions to be more restrictive.
            return False
        # Only after all conditions are met...
        return True

    def _modify(self, **patch):
        """Override modify to check kwargs before request sent to device."""
        if 'state' in patch:
            if patch['state'] != 'user-up' and patch['state'] != 'user-down':
                msg = "The members resource does not support a modify with " \
                      "the value of the 'state' attribute as %s. " \
                      "The accepted values are 'user-up' or " \
                      "'user-down'" % patch['state']
                raise MemberStateModifyUnsupported(msg)
        if 'session' in patch:
            if patch['session'] != 'user-enabled' and patch['session'] != \
                    'user-disabled':
                msg = "The members resource does not support a modify with " \
                      "the value of the 'session' attribute as %s. " \
                      "The accepted values are 'user-enabled' or " \
                      "'user-disabled'" % patch['session']
                raise MemberStateModifyUnsupported(msg)
        super(Members, self)._modify(**patch)



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® LTM data-group submodule.

REST URI
    ``http://localhost/mgmt/tm/ltm/data-group/``

GUI Path
    ``Local Traffic --> iRules --> Data Group List``

REST Kind
    ``tm:ltm:data-group*``
"""

from distutils.version import LooseVersion
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Data_Group(OrganizingCollection):
    def __init__(self, ltm):
        super(Data_Group, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [
            Internals,
            Internal,
            Externals,
            External
        ]


class Internals(Collection):
    def __init__(self, data_groups):
        super(Internals, self).__init__(data_groups)
        self._meta_data['allowed_lazy_attributes'] = [Internal]
        self._meta_data['required_json_kind'] = \
            'tm:ltm:data-group:internal:internalcollectionstate'
        self._meta_data['attribute_registry'] = \
            {'tm:ltm:data-group:internal:internalstate': Internal}
        self._meta_data['uri'] = self._meta_data['uri'].replace('_', '-')


class Internal(Resource):
    def __init__(self, internals):
        super(Internal, self).__init__(internals)
        self._meta_data['required_json_kind'] = \
            'tm:ltm:data-group:internal:internalstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'type', 'records')
        )

    def update(self, **kwargs):
        if 'type' in self.__dict__:
            del self.__dict__['type']
        return self._update(**kwargs)


class Externals(Collection):
    def __init__(self, data_groups):
        super(Externals, self).__init__(data_groups)
        self._meta_data['allowed_lazy_attributes'] = [External]
        self._meta_data['required_json_kind'] =\
            'tm:ltm:data-group:external:externalcollectionstate'
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:data-group:external:externalstate': External}
        self._meta_data['uri'] = self._meta_data['uri'].replace('_', '-')


class External(Resource):
    def __init__(self, externals):
        super(External, self).__init__(externals)
        self._meta_data['required_json_kind'] =\
            'tm:ltm:data-group:external:externalstate'
        self._meta_data['required_creation_parameters'].update(
            ('name', 'externalFileName')
        )

    def update(self, **kwargs):
        if LooseVersion(self._meta_data['bigip']._meta_data['tmos_version']) \
                < LooseVersion('12.0.0'):
            if 'externalFileName' in self.__dict__:
                del self.__dict__['externalFileName']
            if 'type' in self.__dict__:
                del self.__dict__['type']
        return self._update(**kwargs)



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time

from distutils.version import LooseVersion
from f5.bigip.resource import AsmResource
from f5.bigip.resource import Collection
from icontrol.exceptions import iControlUnexpectedHTTPError


from . methods import Methods_s
from . filetypes import Filetypes_s
from . cookies import Cookies_s
from . host_names import Host_Names_s
from . urls import Urls_s
from . parameters import Parameters_s
from . data_guard import Data_Guard
from . blocking_settings import Blocking_Settings
from . whitelist_ips import Whitelist_Ips_s
from . gwt_profiles import Gwt_Profiles_s
from . json_profiles import Json_Profiles_s
from . xml_profiles import Xml_Profiles_s
from . signatures import Signatures_s
from . signature_sets import Signature_Sets_s
from . headers import Headers_s
from . response_pages import Response_Pages_s
from . policy_builder import Policy_Builder
from . history_revisions import History_Revisions_s
from . vulnerability_assessment import Vulnerability_Assessment
from . geolocation_enforcement import Geolocation_Enforcement
from . session_tracking import Session_Tracking
from . session_tracking_status import Session_Tracking_Statuses_s
from . login_pages import Login_Pages_s
from . ip_intelligence import Ip_Intelligence
from . csrf_protection import Csrf_Protection
from . redirection_protection import Redirection_Protection
from . login_enforcement import Login_Enforcement
from . sensitive_parameters import Sensitive_Parameters_s
from . brute_force import Brute_Force_Attack_Preventions_s
from . xml_validation import Xml_Validation_Files_s
from . extractions import Extractions_s
from . vulnerabilities import Vulnerabilities_s
from . navigation_parameters import Navigation_Parameters_s
from . character_sets import Character_Sets_s
from . web_scraping import Web_Scraping
from . audit_logs import Audit_Logs_s
from . suggestions import Suggestions_s
from . plain_text_profiles import Plain_Text_Profiles_s
from . websocket_urls import Websocket_Urls_s
from . general import General


class Policies_s(Collection):
    """BIG-IP® ASM Policies collection."""
    def __init__(self, asm):
        super(Policies_s, self).__init__(asm)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Policy]
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:policystate': Policy
        }


class Policy(AsmResource):
    """BIG-IP® ASM Policies resource."""
    def __init__(self, policies_s):
        super(Policy, self).__init__(policies_s)
        self._meta_data['required_json_kind'] = 'tm:asm:policies:policystate'
        self._meta_data['attribute_registry'] = {
            'tm:asm:policies:methods:methodcollectionstate': Methods_s,
            'tm:asm:policies:filetypes:filetypecollectionstate': Filetypes_s,
            'tm:asm:policies:cookies:cookiecollectionstate': Cookies_s,
            'tm:asm:policies:host-names:host-namecollectionstate': Host_Names_s,
            'tm:asm:policies:urls:urlcollectionstate': Urls_s,
            'tm:asm:policies:parameters:parametercollectionstate': Parameters_s,
            'tm:asm:policies:whitelist-ips:whitelist-ipcollectionstate': Whitelist_Ips_s,
            'tm:asm:policies:gwt-profiles:gwt-profilecollectionstate': Gwt_Profiles_s,
            'tm:asm:policies:json-profiles:json-profilecollectionstate': Json_Profiles_s,
            'tm:asm:policies:xml-profiles:xml-profilecollectionstate': Xml_Profiles_s,
            'tm:asm:policies:signatures:signaturecollectionstate': Signatures_s,
            'tm:asm:policies:signature-sets:signature-setcollectionstate': Signature_Sets_s,
            'tm:asm:policies:headers:headercollectionstate': Headers_s,
            'tm:asm:policies:response-pages:response-pagecollectionstate': Response_Pages_s,
            'tm:asm:policies:history-revisions:history-revisioncollectionstate': History_Revisions_s,
            'tm:asm:policies:vulnerability-assessment:vulnerability-assessmentstate': Vulnerability_Assessment,
            'tm:asm:policies:data-guard:data-guardstate': Data_Guard,
            'tm:asm:policies:geolocation-enforcement:geolocation-enforcementstate': Geolocation_Enforcement,
            'tm:asm:policies:session-tracking:session-awareness-settingsstate': Session_Tracking,
            'tm:asm:policies:session-tracking-statuses:session-tracking-statuscollectionstate': Session_Tracking_Statuses_s,
            'tm:asm:policies:login-pages:login-pagecollectionstate': Login_Pages_s,
            'tm:asm:policies:ip-intelligence:ip-intelligencestate': Ip_Intelligence,
            'tm:asm:policies:csrf-protection:csrf-protectionstate': Csrf_Protection,
            'tm:asm:policies:redirection-protection:redirection-protectionstate': Redirection_Protection,
            'tm:asm:policies:login-enforcement:login-enforcementstate': Login_Enforcement,
            'tm:asm:policies:sensitive-parameters:sensitive-parametercollectionstate': Sensitive_Parameters_s,
            'tm:asm:policies:brute-force-attack-preventions:brute-force-attack-preventioncollectionstate': Brute_Force_Attack_Preventions_s,
            'tm:asm:policies:xml-validation-files:xml-validation-filecollectionstate': Xml_Validation_Files_s,
            'tm:asm:policies:extractions:extractioncollectionstate': Extractions_s,
            'tm:asm:policies:vulnerabilities:vulnerabilitycollectionstate': Vulnerabilities_s,
            'tm:asm:policies:navigation-parameters:navigation-parametercollectionstate': Navigation_Parameters_s,
            'tm:asm:policies:character-sets:character-setcollectionstate': Character_Sets_s,
            'tm:asm:policies:web-scraping:web-scrapingstate': Web_Scraping,
            'tm:asm:policies:audit-logs:audit-logcollectionstate': Audit_Logs_s,
            'tm:asm:policies:suggestions:suggestioncollectionstate': Suggestions_s,
            'tm:asm:policies:plain-text-profiles:plain-text-profilecollectionstate': Plain_Text_Profiles_s,
            'tm:asm:policies:websocket-urls:websocket-urlcollectionstate': Websocket_Urls_s
        }

        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_v) >= LooseVersion('13.0.0'):
            self._meta_data['attribute_registry'].update({
                'tm:asm:policies:general:generalstate': General
            })

        self._set_attr_reg()

    def _set_attr_reg(self):
        """Helper method.

        Appends correct attribute registry, depending on TMOS version

        """
        tmos_v = self._meta_data['bigip']._meta_data['tmos_version']
        attributes = self._meta_data['attribute_registry']
        v12kind = 'tm:asm:policies:blocking-settings:blocking-settingcollectionstate'
        v11kind = 'tm:asm:policies:blocking-settings'
        builderv11 = 'tm:asm:policies:policy-builder:pbconfigstate'
        builderv12 = 'tm:asm:policies:policy-builder:policy-builderstate'
        if LooseVersion(tmos_v) < LooseVersion('12.0.0'):
            attributes[v11kind] = Blocking_Settings
            attributes[builderv11] = Policy_Builder
        else:
            attributes[v12kind] = Blocking_Settings
            attributes[builderv12] = Policy_Builder

    def create(self, **kwargs):
        """Custom creation logic to handle edge cases

        This shouldn't be needed, but ASM has a tendency to raise various errors that
        are painful to handle from a customer point-of-view

        The error itself are described in their exception handler

        To address these failure, we try a number of exception handling cases to catch
        and reliably deal with the error.

        :param kwargs:
        :return:
        """
        for x in range(0, 30):
            try:
                return self._create(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise

    def delete(self, **kwargs):
        """Custom deletion logic to handle edge cases

        This shouldn't be needed, but ASM has a tendency to raise various errors that
        are painful to handle from a customer point-of-view

        The error itself are described in their exception handler

        To address these failure, we try a number of exception handling cases to catch
        and reliably deal with the error.

        :param kwargs:
        :return:
        """
        for x in range(0, 30):
            try:
                return self._delete(**kwargs)
            except iControlUnexpectedHTTPError as ex:
                if self._check_exception(ex):
                    continue
                else:
                    raise

    def _check_exception(self, ex):
        retryable = [
            # iControlUnexpectedHTTPError: 400 Unexpected Error: Bad Request for uri: ...
            # {
            #     "code":400,
            #     "message": "remoteSender:10.0.2.2,method:DELETE ",
            #     "referer":"10.0.2.2",
            #     "restOperationId":103685,
            #     "errorStack":[
            #         "java.util.concurrent.TimeoutException: remoteSender:10.0.2.2, method:DELETE ",
            #         "at com.f5.rest.common.RestWorker.logAndFailExpiredOperation(RestWorker.java:3035)",
            #         "at com.f5.reent.TimeoutException: remoteSender:10.0.2.2, method:DELETE ",
            #         "at com.f5.rest.common.RestWorker.logAndFailExpiredOperation(RestWorker.java:3035)",
            #         "at com.f5.rest.common.RestWorker.checkForExpiredOperations(RestWorker.java:3024)",
            #         "at com.f5.rest.common.RestServer.checkAndExpirePendingWorkerOperations(RestServer.java:1416)",
            #         "at com.f5.rest.common.RestServer.access$400(RestServer.java:45)",
            #         "at com.f5.rest.common.RestServer$4.run(RestServer.java:1389)",
            #         "at com.f5.rest.common.ScheduleTaskManager$2$1.run(ScheduleTaskManager.java:123)",
            #         "at java.util.concurrent.Executors$RunnableAdapter.call(Executors.java:471)",
            #         "at java.util.concurrent.FutureTask.run(FutureTask.java:262)",
            #         "at java.util.concurrent.ScheduledThreadPoolExecutor$ScheduledFutureTask.access$201(ScheduledThreadPoolExecutor.java:178)",
            #         "at java.util.concurrent.ScheduledThreadPoolExecutor$ScheduledFutureTask.run(ScheduledThreadPoolExecutor.java:292)",
            #         "at java.util.concurrent.ThreadPoolExecutor.runWorker(ThreadPoolExecutor.java:1145)",
            #         "at java.util.concurrent.ThreadPoolExecutor$Worker.run(ThreadPoolExecutor.java:615)",
            #         "at java.lang.Thread.run(Thread.java:745)\\n"
            #     ],
            #     "kind":":resterrorresponse"
            # }'
            'TimeoutException',

            # {
            #   "code":400,
            #   "message":"Failed on insert to DCC.FILE_REFERENCES_FOR_DELETION
            #              (DBD::mysql::db do failed: Duplicate entry
            #              \'/ts/var/xml/defense_config/xml_profile_2312\' for key \'filename\')",
            #   "referer":"10.0.2.2",
            #   "restOperationId":142433,
            #   "kind":":resterrorresponse"
            # }
            'Failed on insert to DCC.FILE_REFERENCES_FOR_DELETION',

            # iControlUnexpectedHTTPError: 500 Unexpected Error: Internal Server Error ...
            # {
            #   "code": 500,
            #   "message": "Could not add_signature the Attack Signature.  "
            #              "Failed on insert to PLC.NEGSIG_SET_SIGNATURES "
            #              "(DBD::mysql::db do failed: Lock wait timeout exceeded; "
            #              "try restarting transaction)
            #
            'Lock wait timeout exceeded',
        ]
        if any(x in str(ex) for x in retryable):
            time.sleep(3)
            return True
        elif 'errorStack' in ex:
            stack = ' '.join(ex['errorStack'])
            if any(x in stack for x in retryable):
                time.sleep(3)
                return True
            else:
                return False
        else:
            return False



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® cluster trust-domain submodule

REST URI
    ``http://localhost/mgmt/tm/cm/trust-domain``

GUI Path
    ``Device Management --> Device Trust``

REST Kind
    ``tm:cm:device-group:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource
from f5.sdk_exception import UnsupportedOperation


class Trust_Domains(Collection):
    """BIG-IP® cluster trust-domain collection."""

    def __init__(self, cm):
        super(Trust_Domains, self).__init__(cm)
        self._meta_data['allowed_lazy_attributes'] = [Trust_Domain]
        self._meta_data['attribute_registry'] = \
            {'tm:cm:trust-domain:trust-domainstate': Trust_Domain}


class Trust_Domain(Resource):
    """BIG-IP® cluster trust-domain resource"""
    def __init__(self, trust_domains):
        super(Trust_Domain, self).__init__(trust_domains)
        self._meta_data['required_json_kind'] =\
            'tm:cm:trust-domain:trust-domainstate'
        self._meta_data['required_creation_parameters'].update(('partition',))

    def create(self, **kwargs):
        """Create is not supported for trust domains.

        :raises: :exc:`~f5.BIG-IP.resource.UnsupportedOperation`
        """
        raise UnsupportedOperation(
            "BIG-IP trust domains cannot be created by users")

    def delete(self):
        """Delete is not supported for trust domains.

        :raises: :exc:`~f5.BIG-IP.resource.UnsupportedOperation`
        """
        raise UnsupportedOperation(
            "BIG-IP trust domains cannot be deleted by users")



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Directory: net module: route-domain.

REST URI
    ``https://localhost/mgmt/tm/net/route-domain?ver=11.6.0``

GUI Path
    ``XXX``

REST Kind
    ``tm:net:route-domain:*``
"""


from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Route_Domains(Collection):
    """BIG-IP® Route Domain collection."""
    def __init__(self, net):
        super(Route_Domains, self).__init__(net)
        self._meta_data['allowed_lazy_attributes'] = [Route_Domain]
        self._meta_data['attribute_registry'] =\
            {'tm:net:route-domain:route-domainstate': Route_Domain}


class Route_Domain(Resource):
    """BIG-IP® Route Domain collection."""
    def __init__(self, Route_Domains):
        super(Route_Domain, self).__init__(Route_Domains)
        self._meta_data['read_only_attributes'].append('id')
        self._meta_data['required_json_kind'] =\
            "tm:net:route-domain:route-domainstate"



#!/usr/bin/env python
# Copyright 2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import re
import subprocess
import sys

releaseTagRe = re.compile("^v(\d+\.\d+\.\d+((a|rc|RC)\d+)?$)")
gitDescribeRe = re.compile("^v(\d+\.\d+\.\d+(?:a\d+)?)-(\d+)-g(.*)$")


def runCommand(cmd):
    output = ""

    try:
        p = subprocess.Popen(cmd.split(),
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE
                             )
        (output) = p.communicate()[0]
    except OSError as e:
        print("Execution failed:", e)

    return (output, p.returncode)


def main(args):

    version = "0.0.0"
    releaseNumber = 1
    branch = ""

    # Is this a tagged commit?
    (output, status) = runCommand('git describe --tags --exact-match')

    if status == 0:
        # Yes.
        tag = output.rstrip()

        # Does this tag match the format of tagged releases?
        m = releaseTagRe.match(tag)
        if m:
            version = m.group(1)
        (output, status) = runCommand('git rev-parse --short HEAD')
        if status == 0:
            commit = output.rstrip()
    else:
        # No, use git describe to get information about this commit.
        (output, status) = runCommand('git describe --tags')
        if status == 0:

            m = gitDescribeRe.match(output)
            if m:
                version = m.group(1)
                release = m.group(2)
                commit = m.group(3)

                releaseNumber = releaseNumber + int(release)

    # Get the branch name.
    (output, status) = runCommand('git rev-parse --abbrev-ref HEAD')
    if status == 0:
        branch = output.rstrip()

    results = []
    if args.version:
        results.append(version)
    if args.release:
        results.append(releaseNumber)
    if args.commit:
        results.append(commit)
    if args.branch:
        results.append(branch)

    if len(results):
        print('%s' % ' '.join(map(str, results)))
    else:
        print(version, releaseNumber, commit, branch)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action="store_true")
    parser.add_argument('--release', action="store_true")
    parser.add_argument('--branch', action="store_true")
    parser.add_argument('--commit', action="store_true")

    main(parser.parse_args(sys.argv[1:]))



# coding=utf-8
#
# Copyright 2018 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® SMTP Server module

REST URI
    ``http://localhost/mgmt/sys/smtp-server/``

GUI Path
    ``Systems > Configuration > Device > SMTP``

REST Kind
    ``tm:sys:smtp-server:*``
"""

from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Smtp_Servers(Collection):
    def __init__(self, sys):
        super(Smtp_Servers, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [Smtp_Server]
        self._meta_data['attribute_registry'] = {
            'tm:sys:smtp-server:smtp-serverstate': Smtp_Server
        }


class Smtp_Server(Resource, ExclusiveAttributesMixin):
    def __init__(self, smtp_servers):
        super(Smtp_Server, self).__init__(smtp_servers)
        self._meta_data['required_json_kind'] = 'tm:sys:smtp-server:smtp-serverstate'
        self._meta_data['exclusive_attributes'].append(('authenticationDisabled', 'authenticationEnabled'))

    def _endis_attrs(self):
        if 'authenticationDisabled' in self.__dict__:
            self.__dict__['authenticationEnabled'] = not self.__dict__['authenticationDisabled']
        if 'authenticationEnabled' in self.__dict__:
            self.__dict__['authenticationDisabled'] = not self.__dict__['authenticationEnabled']
        return None

    def create(self, **kwargs):
        inst = self._create(**kwargs)
        inst._endis_attrs()
        return inst

    def load(self, **kwargs):
        kwargs = self._reduce_boolean_pair(kwargs, 'authenticationEnabled', 'authenticationDisabled')
        inst = self._load(**kwargs)
        inst._endis_attrs()
        return inst

    def refresh(self, **kwargs):
        kwargs = self._reduce_boolean_pair(kwargs, 'authenticationEnabled', 'authenticationDisabled')
        self._refresh(**kwargs)
        self._endis_attrs()
        return self

    def update(self, **kwargs):
        if 'authenticationEnabled' in kwargs or 'authenticationDisabled' in kwargs:
            self.__dict__.pop('authenticationEnabled')
            self.__dict__.pop('authenticationDisabled')
        kwargs = self._reduce_boolean_pair(kwargs, 'authenticationEnabled', 'authenticationDisabled')
        self.__dict__ = self._reduce_boolean_pair(self.__dict__, 'authenticationEnabled', 'authenticationDisabled')
        self._update(**kwargs)
        self._endis_attrs()



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""BIG-IP® system dns module

REST URI
    ``http://localhost/mgmt/tm/sys/dns``

tmsh Path
    ``sys --> httpd --> all-properties``

GUI Path
    ``various``

REST Kind
    ``tm:sys:httpd:*``
"""

from f5.bigiq.resource import UnnamedResource


class Httpd(UnnamedResource):
    """BIG-IP® system HTTPD unnamed resource

        .. note::

        This is an unnamed resource so it has no ~Partition~Name pattern
        at the end of its URI.
    """
    def __init__(self, sys):
        super(Httpd, self).__init__(sys)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['required_json_kind'] = 'tm:sys:httpd:httpdstate'



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""iWorkflow® system setup API

REST URI
    ``http://localhost/mgmt/shared/system/easy-setup``

REST Kind
    none
"""

from f5.iworkflow.resource import UnnamedResource


class Easy_Setup(UnnamedResource):
    def __init__(self, config):
        super(Easy_Setup, self).__init__(config)
        self._meta_data['required_load_parameters'] = set()



# coding=utf-8
# Copyright 2016 F5 Networks Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#

'''Class for managing a DeviceGroup for a set of BIG-IP® devices

Managing a device group for clustering is an event-driven process. Please
use the methods here to control that process. The fundamental idea is that
any action should have an observable outcome. Adding a device to the device
group should have a consequence for each member of the device group,
including the newly added member.

Examples:

There are two major use-cases here:

    * Manage an existing device group:

        list_of_bigips = [ManagementRoot(...), ManagementRoot(...)]
        device_group = DeviceGroup(
                            devices=list_of_bigips,
                            device_group_name='my_cluster',
                            device_group_type='sync-failover',
                            device_group_partition='Common'
                        )
        device_group.ensure_all_devices_in_sync()

    * Create a new device group and manage it:

        list_of_bigips = [ManagementRoot(...), ManagementRoot(...)]
        device_group = DeviceGroup()
        device_group.create(
                        devices=list_of_bigips,
                        device_group_name='my_cluster',
                        device_group_type='sync-failover',
                        device_group_partition='Common'
                    )
        device_group.ensure_all_devices_in_sync()

Methods:

    * create -- create a device group from a list of devices
    * teardown -- teardown a device group, but leave the trust domain intact
    * validate -- ensure a device group is in the proper state based on inputs
    * manage_extant -- manage an existing device group

'''

from f5.multi_device.exceptions import DeviceGroupNotSupported
from f5.multi_device.exceptions import MissingRequiredDeviceGroupParameter
from f5.multi_device.exceptions import UnexpectedDeviceGroupDevices
from f5.multi_device.exceptions import UnexpectedDeviceGroupState
from f5.multi_device.exceptions import UnexpectedDeviceGroupType

from f5.multi_device.utils import get_device_info
from f5.multi_device.utils import pollster


class DeviceGroup(object):
    '''Class to manage device service group

    For the non-public methods, there are a few flavors of behavior:
    get, check, and ensure. A 'get' retrieves some info from the device
    without any assumptions about that info. A 'check' will assert a device's
    info is as expected. An 'ensure' method often does one or more of the
    above and also may take some other action to enforce the expected state,
    such as syncing config.

    The pollster is used heavliy here for 'check' and 'get' methods, since we
    are often waiting for the device or devices to respond to some action.

    Example:

        * dg = self._get_device_group()
        * self._check_all_devices_in_sync()
        * self.ensure_all_devices_in_sync()

    '''

    available_types = ['sync-failover', 'sync-only']
    sync_status_entry = 'https://localhost/mgmt/tm/cm/sync-status/0'

    def __init__(self, **kwargs):
        '''Initialize a device group manager.

        '''

        if kwargs:
            self.manage_extant(**kwargs)

    def _set_attributes(self, **kwargs):
        '''Set instance attributes based on kwargs

        :param kwargs: dict -- kwargs to set as attributes
        '''

        try:
            self.devices = kwargs['devices'][:]
            self.name = kwargs['device_group_name']
            self.type = kwargs['device_group_type']
            self.partition = kwargs['device_group_partition']
        except KeyError as ex:
            raise MissingRequiredDeviceGroupParameter(ex)

    def validate(self, **kwargs):
        '''Validate device group state among given devices.

        :param kwargs: dict -- keyword args of device group information
        :raises: UnexpectedDeviceGroupType, UnexpectedDeviceGroupDevices
        '''

        self._set_attributes(**kwargs)
        self._check_type()
        self.dev_group_uri_res = self._get_device_group(self.devices[0])
        if self.dev_group_uri_res.type != self.type:
            msg = 'Device group type found: %r does not match expected ' \
                'device group type: %r' % (
                    self.dev_group_uri_res.type, self.type
                )
            raise UnexpectedDeviceGroupType(msg)
        queried_device_names = self._get_device_names_in_group()
        given_device_names = []
        for device in self.devices:
            device_name = get_device_info(device).name
            given_device_names.append(device_name)
        if sorted(queried_device_names) != sorted(given_device_names):
            msg = 'Given devices does not match queried devices.'
            raise UnexpectedDeviceGroupDevices(msg)
        self.ensure_all_devices_in_sync()

    def _check_type(self):
        '''Check that the device group type is correct.

        :raises: DeviceGroupOperationNotSupported, DeviceGroupNotSupported
        '''

        if self.type not in self.available_types:
            msg = 'Unsupported cluster type was given: %s' % self.type
            raise DeviceGroupNotSupported(msg)
        elif self.type == 'sync-only' and self.name != 'device_trust_group':
            msg = "Management of sync-only device groups only supported for " \
                "built-in device group named 'device_trust_group'"
            raise DeviceGroupNotSupported(msg)

    def manage_extant(self, **kwargs):
        self.validate(**kwargs)

    def create(self, **kwargs):
        '''Create the device service cluster group and add devices to it.'''

        self._set_attributes(**kwargs)
        self._check_type()
        pollster(self._check_all_devices_in_sync)()
        dg = self.devices[0].tm.cm.device_groups.device_group
        dg.create(name=self.name, partition=self.partition, type=self.type)
        for device in self.devices:
            self._add_device_to_device_group(device)
            device.tm.sys.config.exec_cmd('save')
        self.ensure_all_devices_in_sync()

    def teardown(self):
        '''Teardown device service cluster group.'''

        self.ensure_all_devices_in_sync()
        for device in self.devices:
            self._delete_device_from_device_group(device)
            self._sync_to_group(device)
            pollster(self._ensure_device_active)(device)
            self.ensure_all_devices_in_sync()
        dg = pollster(self._get_device_group)(self.devices[0])
        dg.delete()
        pollster(self._check_devices_active_licensed)()
        pollster(self._check_all_devices_in_sync)()

    def _get_device_names_in_group(self):
        '''_get_device_names_in_group

        :returns: list -- list of device names in group
        '''

        device_names = []
        dg = pollster(self._get_device_group)(self.devices[0])
        members = dg.devices_s.get_collection()
        for member in members:
            member_name = member.name.replace('/%s/' % self.partition, '')
            device_names.append(member_name)
        return device_names

    def _get_device_group(self, device):
        '''Get the device group through a device.

        :param device: bigip object -- device
        :returns: tm.cm.device_groups.device_group object
        '''

        return device.tm.cm.device_groups.device_group.load(
            name=self.name, partition=self.partition
        )

    def _check_devices_active_licensed(self):
        '''All devices should be in an active/licensed state.

        :raises: UnexpectedClusterState
        '''

        if len(self._get_devices_by_activation_state('active')) != \
                len(self.devices):
            msg = "One or more devices not in 'Active' and licensed state."
            raise UnexpectedDeviceGroupState(msg)

    def _add_device_to_device_group(self, device):
        '''Add device to device service cluster group.

        :param device: bigip object -- device to add to group
        '''

        device_name = get_device_info(device).name
        dg = pollster(self._get_device_group)(device)
        dg.devices_s.devices.create(name=device_name, partition=self.partition)
        pollster(self._check_device_exists_in_device_group)(device_name)

    def _check_device_exists_in_device_group(self, device_name):
        '''Check whether a device exists in the device group

        :param device: ManagementRoot object -- device to look for
        '''

        dg = self._get_device_group(self.devices[0])
        dg.devices_s.devices.load(name=device_name, partition=self.partition)

    def _delete_device_from_device_group(self, device):
        '''Remove device from device service cluster group.

        :param device: ManagementRoot object -- device to delete from group
        '''

        device_name = get_device_info(device).name
        dg = pollster(self._get_device_group)(device)
        device_to_remove = dg.devices_s.devices.load(
            name=device_name, partition=self.partition
        )
        device_to_remove.delete()

    def _ensure_device_active(self, device):
        '''Ensure a single device is in an active state

        :param device: ManagementRoot object -- device to inspect
        :raises: UnexpectedClusterState
        '''

        act = device.tm.cm.devices.device.load(
            name=get_device_info(device).name,
            partition=self.partition
        )
        if act.failoverState != 'active':
            msg = "A device in the cluster was not in the 'Active' state."
            raise UnexpectedDeviceGroupState(msg)

    def _sync_to_group(self, device):
        '''Sync the device to the cluster group

        :param device: bigip object -- device to sync to group
        '''

        config_sync_cmd = 'config-sync to-group %s' % self.name
        device.tm.cm.exec_cmd('run', utilCmdArgs=config_sync_cmd)

    def ensure_all_devices_in_sync(self):
        """Ensure all devices have 'In Sync' status are sync is done."""

        self._sync_to_group(self.devices[0])
        pollster(self._check_all_devices_in_sync)()

    def _check_all_devices_in_sync(self):
        '''Wait until all devices have failover status of 'In Sync'.

        :raises: UnexpectedClusterState
        '''

        if len(self._get_devices_by_failover_status('In Sync')) != \
                len(self.devices):
            msg = "Expected all devices in group to have 'In Sync' status."
            raise UnexpectedDeviceGroupState(msg)

    def _get_devices_by_failover_status(self, status):
        '''Get a list of bigips by failover status.

        :param status: str -- status to filter the returned list of devices
        :returns: list -- list of devices that have the given status
        '''

        devices_with_status = []
        for device in self.devices:
            if (self._check_device_failover_status(device, status)):
                devices_with_status.append(device)
        return devices_with_status

    def _check_device_failover_status(self, device, status):
        '''Determine if a device has a specific failover status.

        :param status: str -- status to check against
        :returns: bool -- True is it has status, False otherwise
        '''

        sync_status = device.tm.cm.sync_status
        sync_status.refresh()
        current_status = (sync_status.entries[self.sync_status_entry]
                          ['nestedStats']['entries']['status']
                          ['description'])
        if status == current_status:
            return True
        return False

    def _get_devices_by_activation_state(self, state):
        '''Get a list of bigips by activation statue.

        :param state: str -- state to filter the returned list of devices
        :returns: list -- list of devices that are in the given state
        '''

        devices_with_state = []
        for device in self.devices:
            act = device.tm.cm.devices.device.load(
                name=get_device_info(device).name,
                partition=self.partition
            )
            if act.failoverState == state:
                devices_with_state.append(device)
        return devices_with_state



# coding=utf-8
#
# Copyright 2015 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® System Software module

REST URI
    ``http://localhost/mgmt/tm/sys/software``

GUI Path
    ``System``

REST Kind
    ``tm:sys:software:*``
"""

from f5.bigip.resource import OrganizingCollection
from f5.bigip.tm.sys.software.hotfix import Hotfix_s
from f5.bigip.tm.sys.software.image import Images
from f5.bigip.tm.sys.software.update import Update
from f5.bigip.tm.sys.software.volume import Volumes


class Software(OrganizingCollection):
    def __init__(self, tm):
        super(Software, self).__init__(tm)
        self._meta_data['allowed_lazy_attributes'] = [
            Hotfix_s,
            Images,
            Update,
            Volumes
        ]

    def __getattribute__(self, name):
        if name == 'update':
            return Update(self)
        return super(Software, self).__getattribute__(name)



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.mixins import ExclusiveAttributesMixin
from f5.bigip.resource import UnnamedResource


class Add_To_Trust(UnnamedResource, ExclusiveAttributesMixin,
                   CommandExecutionMixin):
    """BIG-IP® Add-To-Trust resource

    Use this object to set or overwrite device trust

    """

    def __init__(self, cm):
        super(Add_To_Trust, self).__init__(cm)
        self._meta_data['exclusive_attributes'].append(
            ('caDevice', 'nonCaDevice'))
        self._meta_data['required_command_parameters'].update(
            ('name', 'device', 'deviceName', 'username', 'password'))
        self._meta_data['required_json_kind'] = \
            'tm:cm:add-to-trust:runstate'
        self._meta_data['allowed_commands'].append('run')
        self._meta_data['minimum_version'] = '12.0.0'


class Remove_From_Trust(UnnamedResource, CommandExecutionMixin):
    """BIG-IP®« Remove-From-Trust resource

    Use this object to remove device trust

    .. note::
        This will only remove trust setting on a single BIG-IP®.
        Full trust removal requires that the operation is
        carried out on both target devices

    """

    def __init__(self, cm):
        super(Remove_From_Trust, self).__init__(cm)
        self._meta_data['required_command_parameters'].update(
            ('deviceName', 'name'))
        self._meta_data['required_json_kind'] = \
            'tm:cm:remove-from-trust:runstate'
        self._meta_data['allowed_commands'].append('run')
        self._meta_data['minimum_version'] = '12.0.0'



# coding=utf-8
#
#  Copyright 2014-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""iWorkflow® users module.

REST URI
    ``http://localhost/mgmt/cm/shared/authz/users``

REST Kind
    ``shared:authz:users:*``
"""

from f5.iworkflow.resource import Collection
from f5.iworkflow.resource import Resource


class Users_s(Collection):
    def __init__(self, authz):
        super(Users_s, self).__init__(authz)
        self._meta_data['required_json_kind'] = \
            'shared:authz:users:userscollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [User]
        self._meta_data['attribute_registry'] = {
            'shared:authz:users:usersworkerstate': User
        }


class User(Resource):
    def __init__(self, users_s):
        super(User, self).__init__(users_s)
        self._meta_data['required_load_parameters'] = {'name', }
        # There's a confirmation password parameter here (password2) but it is
        # not needed by the API. The web UI, however, requires it.
        self._meta_data['required_creation_parameters'] = \
            {'name', 'password'}
        self._meta_data['required_json_kind'] = \
            'shared:authz:users:usersworkerstate'



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® system config module

REST URI
    ``http://localhost/mgmt/tm/sys/config``

GUI Path
    N/A

REST Kind
    ``tm:sys:config:*``
"""
from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Crypto(OrganizingCollection):
    def __init__(self, sys):
        super(Crypto, self).__init__(sys)
        self._meta_data['allowed_lazy_attributes'] = [
            Certs,
            Keys
        ]


class Keys(Collection, CommandExecutionMixin):
    """BIG-IP® Crypto key collection


        note::
        This collection supports install command.
        Given the fact that we will be expecting hyphen
        parameters, the function will need to utilize
        variable keyword argument syntax. In other words
        define a dictionary with the arbitrary keys and
        then pass it as in the form **foo into the method
        call. e.g.

        param_set ={'from-local-file': FOOPATH, 'name': 'FOOKEY'}
        bigip.tm.sys.crypto.keys.exec_cmd('install', **param_set)


    """
    def __init__(self, crypto):
        super(Keys, self).__init__(crypto)
        self._meta_data['allowed_lazy_attributes'] = [Key]
        self._meta_data['allowed_commands'].append('install')
        self._meta_data['attribute_registry'] =\
            {'tm:sys:crypto:key:keystate': Key}


class Key(Resource):
    """BIG-IP® Crypto key resource"""
    def __init__(self, keys):
        super(Key, self).__init__(keys)
        self._meta_data['required_json_kind'] = 'tm:sys:crypto:key:keystate'


class Certs(Collection, CommandExecutionMixin):
    """BIG-IP® Crypto cert collection


        note::
        This collection supports install command.
        Given the fact that we will be expecting hyphen
        parameters, the function will need to utilize
        variable keyword argument syntax. In other words
        define a dictionary with the arbitrary keys and
        then pass it as in the form **foo into the method
        call. e.g.

        param_set ={'from-local-file': FOOPATH, 'name': 'FOOCERT'}
        bigip.tm.sys.crypto.certs.exec_cmd('install', **param_set)


    """
    def __init__(self, crypto):
        super(Certs, self).__init__(crypto)
        self._meta_data['allowed_lazy_attributes'] = [Cert]
        self._meta_data['allowed_commands'].append('install')
        self._meta_data['attribute_registry'] =\
            {'tm:sys:crypto:cert:certstate': Cert}


class Cert(Resource):
    """BIG-IP® Crypto cert resource"""
    def __init__(self, certs):
        super(Cert, self).__init__(certs)
        self._meta_data['required_json_kind'] = 'tm:sys:crypto:cert:certstate'



# coding=utf-8
#
# Copyright 2015-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Advanced Firewall Manager™ (AFM®) module.

REST URI
    ``http://localhost/mgmt/tm/security/firewall``

GUI Path
    ``Security --> Network Firewall``

REST Kind
    ``tm:security:firewall:*``
"""
from distutils.version import LooseVersion
from f5.bigip.mixins import CheckExistenceMixin
from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource
from f5.bigip.resource import UnnamedResource
from f5.sdk_exception import NonExtantFirewallRule


class Firewall(OrganizingCollection):
    """BIG-IP® AFM® Firewall organizing collection."""

    def __init__(self, security):
        super(Firewall, self).__init__(security)
        self._meta_data['allowed_lazy_attributes'] = [
            Address_Lists,
            Port_Lists,
            Rule_Lists,
            Policy_s,
            Global_Rules]


class Address_Lists(Collection):
    """BIG-IP® AFM® Address List collection"""
    def __init__(self, firewall):
        super(Address_Lists, self).__init__(firewall)
        self._meta_data['allowed_lazy_attributes'] = [Address_List]
        self._meta_data['attribute_registry'] = \
            {'tm:security:firewall:address-list:address-liststate':
                Address_List}


class Address_List(Resource):
    """BIG-IP® Address List resource"""
    def __init__(self, address_lists):
        super(Address_List, self).__init__(address_lists)
        self._meta_data['required_json_kind'] = \
            'tm:security:firewall:address-list:address-liststate'
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_load_parameters'].update(('partition',))
        self.tmos_ver = self._meta_data['bigip'].tmos_version
        if LooseVersion(self.tmos_ver) < LooseVersion('12.0.0'):
            self._meta_data['minimum_additional_parameters'] = {
                'addressLists', 'addresses', 'geo'}
        else:
            self._meta_data['minimum_additional_parameters'] = {
                'addressLists', 'addresses', 'geo', 'fqdns'}


class Port_Lists(Collection):
    """BIG-IP® AFM® Port List collection"""
    def __init__(self, firewall):
        super(Port_Lists, self).__init__(firewall)
        self._meta_data['allowed_lazy_attributes'] = [Port_List]
        self._meta_data['attribute_registry'] = \
            {'tm:security:firewall:port-list:port-liststate':
                Port_List}


class Port_List(Resource):
    """BIG-IP® Port List resource"""
    def __init__(self, port_lists):
        super(Port_List, self).__init__(port_lists)
        self._meta_data['required_json_kind'] = \
            'tm:security:firewall:port-list:port-liststate'
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_load_parameters'].update(('partition',))
        self._meta_data['minimum_additional_parameters'] = {'ports',
                                                            'portLists'}


class Rule_Lists(Collection):
    """BIG-IP® AFM® Rule List collection"""
    def __init__(self, firewall):
        super(Rule_Lists, self).__init__(firewall)
        self._meta_data['allowed_lazy_attributes'] = [Rule_List]
        self._meta_data['attribute_registry'] = \
            {'tm:security:firewall:rule-list:rule-liststate':
                Rule_List}


class Rule_List(Resource):
    """BIG-IP® Rule List resource"""
    def __init__(self, rule_lists):
        super(Rule_List, self).__init__(rule_lists)
        self._meta_data['required_json_kind'] = \
            'tm:security:firewall:rule-list:rule-liststate'
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_load_parameters'].update(('partition',))
        self._meta_data['allowed_lazy_attributes'] = [Rules_s]
        self._meta_data['attribute_registry'] = \
            {'tm:security:firewall:rule-list:rules:rulescollectionstate':
                Rules_s}


class Rules_s(Collection):
    """BIG-IP® AFM® Rules sub-collection."""
    def __init__(self, rule_list):
        super(Rules_s, self).__init__(rule_list)
        self._meta_data['required_json_kind'] = \
            'tm:security:firewall:rule-list:rules:rulescollectionstate'
        self._meta_data['allowed_lazy_attributes'] = [Rule]
        self._meta_data['attribute_registry'] = \
            {'tm:security:firewall:rule-list:rules:rulesstate':
                Rule}


class Rule(Resource, CheckExistenceMixin):
    """BIG-IP® AFM® Rule resource.


    NOTE:: The 'place-before' and 'place-after' attribute are
        mandatory but cannot be present with one another. Those attributes
        will not be visible when the class is created, they exist for the
        sole purpose of rule ordering in the BIGIP. The ordering of the
        rules corresponds to the index in the 'items' of the Rules_s
        sub-collection.
    """
    def __init__(self, rules_s):
        super(Rule, self).__init__(rules_s)
        self._meta_data['required_json_kind'] = \
            'tm:security:firewall:rule-list:rules:rulesstate'
        self._meta_data['required_creation_parameters'].update(('action',))
        self._meta_data['exclusive_attributes'].append(
            ('place-after', 'place-before'))
        self._meta_data['minimum_additional_parameters'] = {'place-before',
                                                            'place-after'}
        self.tmos_ver = self._meta_data['bigip'].tmos_version

    def update(self, **kwargs):
        """We need to implement the custom exclusive parameter check."""
        self._check_exclusive_parameters(**kwargs)
        return super(Rule, self)._update(**kwargs)

    def modify(self, **kwargs):
        """We need to implement the custom exclusive parameter check."""
        self._check_exclusive_parameters(**kwargs)
        return super(Rule, self)._modify(**kwargs)

    def load(self, **kwargs):
        """Custom load method to address issue in 11.6.0 Final,

        where non existing objects would be True.
        """
        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._load_11_6(**kwargs)
        else:
            return super(Rule, self)._load(**kwargs)

    def _load_11_6(self, **kwargs):
        """Must check if rule actually exists before proceeding with load."""
        if self._check_existence_by_collection(self._meta_data['container'],
                                               kwargs['name']):
            return super(Rule, self)._load(**kwargs)
        msg = 'The application resource named, {}, does not exist on the ' \
              'device.'.format(kwargs['name'])
        raise NonExtantFirewallRule(msg)

    def exists(self, **kwargs):
        """Some objects when deleted still return when called by their

        direct URI, this is a known issue in 11.6.0.
        """

        if LooseVersion(self.tmos_ver) == LooseVersion('11.6.0'):
            return self._exists_11_6(**kwargs)
        else:
            return super(Rule, self)._load(**kwargs)

    def _exists_11_6(self, **kwargs):
        """Check rule existence on device."""

        return self._check_existence_by_collection(
            self._meta_data['container'], kwargs['name'])


class Policy_s(Collection):
    """BIG-IP® AFM® Policy collection"""
    def __init__(self, firewall):
        super(Policy_s, self).__init__(firewall)
        self._meta_data['allowed_lazy_attributes'] = [Policy]
        self._meta_data['attribute_registry'] = \
            {'tm:security:firewall:policy:policystate':
                Policy}


class Policy(Resource):
    """BIG-IP® AFM® Policy resource"""
    def __init__(self, policy_s):
        super(Policy, self).__init__(policy_s)
        self._meta_data['required_json_kind'] = \
            'tm:security:firewall:policy:policystate'
        self._meta_data['required_creation_parameters'].update(('partition',))
        self._meta_data['required_load_parameters'].update(('partition',))
        self._meta_data['allowed_lazy_attributes'] = [Rules_s]
        self._meta_data['attribute_registry'] = \
            {'tm:security:firewall:rule-list:rules:rulescollectionstate':
                Rules_s}


class Global_Rules(UnnamedResource):
    """BIG-IP® AFM® Global Rules resource"""
    def __init__(self, global_rules):
        super(Global_Rules, self).__init__(global_rules)
        self._meta_data['required_json_kind'] = \
            'tm:security:firewall:global-rules:global-rulesstate'



# coding=utf-8
#
#  Copyright 2014-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Global Traffic Manager™ (GTM®) pool module.

REST URI
    ``http://localhost/mgmt/tm/gtm/server``

GUI Path
    ``DNS --> GSLB : Servers``

REST Kind
    ``tm:gtm:server:*``
"""

from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Servers(Collection):
    """BIG-IP® GTM server collection"""
    def __init__(self, gtm):
        super(Servers, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [Server]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:server:serverstate': Server}


class Server(Resource):
    """BIG-IP® GTM server resource"""
    def __init__(self, servers):
        super(Server, self).__init__(servers)
        self._meta_data['required_json_kind'] = 'tm:gtm:server:serverstate'
        self._meta_data['required_creation_parameters'].update(
            ('datacenter', 'addresses'))
        self._meta_data['attribute_registry'] = {
            'tm:gtm:server:virtual-servers:virtual-serverscollectionstate':
                Virtual_Servers_s
        }


class Virtual_Servers_s(Collection):
    """BIG-IP® GTM virtual server sub-collection"""
    def __init__(self, server):
        super(Virtual_Servers_s, self).__init__(server)
        self._meta_data['allowed_lazy_attributes'] = [Virtual_Server]
        self._meta_data['required_json_kind'] = \
            'tm:gtm:server:virtual-servers:virtual-serverscollectionstate'
        self._meta_data['attribute_registry'] = {
            'tm:gtm:server:virtual-servers:virtual-serversstate':
                Virtual_Server}


class Virtual_Server(Resource):
    """BIG-IP® GTM virtual server resource"""
    def __init__(self, virtual_servers_s):
        super(Virtual_Server, self).__init__(virtual_servers_s)
        self._meta_data['required_creation_parameters'].update((
            'destination',))
        self._meta_data['required_json_kind'] = \
            'tm:gtm:server:virtual-servers:virtual-serversstate'



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""BIG-IP® system config module

REST URI
    ``http://localhost/mgmt/tm/sys/ucs`

GUI Path
    N/A

REST Kind
    ``tm:sys:ucs:*``
"""

from f5.bigip.mixins import CommandExecutionMixin
from f5.bigip.resource import UnnamedResource
from requests.exceptions import HTTPError


class Ucs(UnnamedResource, CommandExecutionMixin):
    """BIG-IP® system UCS resource

        .. note::
        Given the fact that 11.6.0 UCS via rest is
        broken, this feature will be supported in 12.0.0 and above.
        Listing of installed UCS has been fixed in 12.1.0.
        This resource is a collection which does not allow listing
        of each ucs as a resource. 'Items' attribute of the loaded object
        is used to access the list of installed UCS files.

        Caveat:
        Loading UCS will result in ICRD restarting, therefore
        due to ID476518 502 Bad Gateway is generated, this is
        working as intended, at least until some architecture
        changes have been made.


    """
    def __init__(self, sys):
        super(Ucs, self).__init__(sys)
        self._meta_data['required_load_parameters'] = set()
        self._meta_data['allowed_commands'].extend(['load', 'save'])
        self._meta_data['required_json_kind'] = ''
        self._meta_data['minimum_version'] = '12.0.0'

    def exec_cmd(self, command, **kwargs):
        """Due to ID476518 the load command need special treatment."""
        self._is_allowed_command(command)
        self._check_command_parameters(**kwargs)

        if command == 'load':
            kwargs['command'] = command
            self._check_exclusive_parameters(**kwargs)
            requests_params = self._handle_requests_params(kwargs)
            session = self._meta_data['bigip']._meta_data['icr_session']
            try:

                session.post(
                    self._meta_data['uri'], json=kwargs, **requests_params)

            except HTTPError as err:
                if err.response.status_code != 502:
                    raise
                return

        else:
            return self._exec_cmd(command, **kwargs)

    def load(self, **kwargs):
        """Method to list the UCS on the system

        Since this is only fixed in 12.1.0 and up
        we implemented version check here
        """

        # Check if we are using 12.1.0 version or above when using this method
        self._is_version_supported_method('12.1.0')
        newinst = self._stamp_out_core()
        newinst._refresh(**kwargs)

        return newinst



# coding=utf-8
#
"""Classes and functions for configuring BIG-IP"""
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import OrganizingCollection

from f5.bigip.tm.asm import Asm
from f5.bigip.tm.auth import Auth
from f5.bigip.tm.cm import Cm
from f5.bigip.tm.gtm import Gtm
from f5.bigip.tm.ltm import Ltm
from f5.bigip.tm.net import Net
from f5.bigip.tm.security import Security
from f5.bigip.tm.shared import Shared
from f5.bigip.tm.sys import Sys
from f5.bigip.tm.transaction import Transactions
from f5.bigip.tm.util import Util
from f5.bigip.tm.vcmp import Vcmp


class Tm(OrganizingCollection):
    """An organizing collection for TM resources."""
    def __init__(self, bigip):
        super(Tm, self).__init__(bigip)
        self._meta_data['allowed_lazy_attributes'] = [
            Asm,
            Auth,
            Cm,
            Gtm,
            Ltm,
            Net,
            Security,
            Shared,
            Sys,
            Transactions,
            Util,
            Vcmp
        ]



# coding=utf-8
#
#  Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Directory: ltm module: virtual-address.

REST URI
    ``https://localhost/mgmt/tm/ltm/virtual-address?ver=11.6.0``

GUI Path
    ``Local Traffic Manager --> Virtual Servers --> Virtual Address List``

REST Kind
    ``tm:ltm:virtual-address:*``
"""


from f5.bigip.resource import Collection
from f5.bigip.resource import Resource


class Virtual_Address_s(Collection):
    """BIG-IP® LTM virtual address collection."""
    def __init__(self, ltm):
        super(Virtual_Address_s, self).__init__(ltm)
        self._meta_data['allowed_lazy_attributes'] = [Virtual_Address]
        self._meta_data['attribute_registry'] =\
            {'tm:ltm:virtual-address:virtual-addressstate': Virtual_Address}


class Virtual_Address(Resource):
    """BIG-IP® LTM virtual address resource."""
    def __init__(self, Virtual_Address_s):
        super(Virtual_Address, self).__init__(Virtual_Address_s)
        self._meta_data['required_json_kind'] =\
            "tm:ltm:virtual-address:virtual-addressstate"
        self._meta_data['read_only_attributes'].append('address')



# coding=utf-8
#
# Copyright 2015-2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip.resource import AsmResource
from f5.bigip.resource import Collection
from f5.sdk_exception import UnsupportedOperation


class Signature_Statuses_s(Collection):
    """BIG-IP® ASM Signature Statuses collection."""
    def __init__(self, asm):
        super(Signature_Statuses_s, self).__init__(asm)
        self._meta_data['object_has_stats'] = False
        self._meta_data['allowed_lazy_attributes'] = [Signature_Status]
        self._meta_data['attribute_registry'] = {
            'tm:asm:signature-statuses:signature-statusstate':
                Signature_Status}


class Signature_Status(AsmResource):
    """BIG-IP® ASM Signature Status resource"""
    def __init__(self, signature_statuses_s):
        super(Signature_Status, self).__init__(signature_statuses_s)
        self._meta_data['required_json_kind'] =\
            'tm:asm:signature-statuses:signature-statusstate'

    def create(self, **kwargs):
        """Create is not supported for Signature Status resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the create method" % self.__class__.__name__
        )

    def delete(self, **kwargs):
        """Delete is not supported for Signature Status resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the delete method" % self.__class__.__name__
        )

    def modify(self, **kwargs):
        """Modify is not supported for Signature Status resource

                :raises: UnsupportedOperation
        """
        raise UnsupportedOperation(
            "%s does not support the modify method" % self.__class__.__name__
        )



# coding=utf-8
#
#  Copyright 2014-2017 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""BIG-IP® Global Traffic Manager™ (GTM®) pool module.

REST URI
    ``http://localhost/mgmt/tm/gtm/wideip``

GUI Path
    ``DNS --> GSLB : Wide IPs``

REST Kind
    ``tm:gtm:pools:*``
"""

from distutils.version import LooseVersion

from f5.bigip.resource import Collection
from f5.bigip.resource import OrganizingCollection
from f5.bigip.resource import Resource


class Wideips(object):
    """BIG-IP® GTM WideIP factory

       The GTM WideIP factory object is used to provide a consistent user
       experience across the SDK, while supporting the breaking changes in
       the BIG-IP APIs.

       Between the 11.x and 12.x releases of BIG-IP, the REST endpoint for
       WideIP changed from a Collection to an OrganizingCollection. The result
       is that breaking changes occurred in the API.

       This breakage led to a discussion on naming conventions because there
       appeared to be a conflict now. For example, depending on your version,
       only one of the following would work.

       For 11.x,

           tm.gtm.wideips.wideip.load(name='foo')

       For 12.x,

           tm.gtm.wideips.a_s.a.load(name='foo')

       but not both. To stick with a consistent user experience, we decided
       to override the __new__ method to support both examples above. The SDK
       automatically detects which one to use based on the BIG-IP you are
       communicating with.
       """
    def __new__(cls, container):
        tmos_v = container._meta_data['bigip']._meta_data['tmos_version']
        if LooseVersion(tmos_v) < LooseVersion('12.0.0'):
            return WideipCollection(container)
        else:
            return WideipOrganizingCollection(container)


class WideipOrganizingCollection(OrganizingCollection):
    """v12.x GTM WideIP is an OC."""
    def __init__(self, gtm):
        self.__class__.__name__ = 'Wideip'
        super(WideipOrganizingCollection, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [
            A_s,
            Aaaas,
            Cnames,
            Mxs,
            Naptrs,
            Srvs,
            ]


class A_s(Collection):
    """v12.x BIG-IP® A wideip collection.

        Class name needed changing due to
        'as' being reserved python keyword
    """
    def __init__(self, wideip):
        super(A_s, self).__init__(wideip)
        self._meta_data['allowed_lazy_attributes'] = [A]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:wideip:a:astate': A}


class A(Resource):
    """v12.x BIG-IP®A wideip resource"""
    def __init__(self, _as):
        super(A, self).__init__(_as)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:wideip:a:astate'


class Aaaas(Collection):
    """v12.x BIG-IP® AAAA wideip collection"""
    def __init__(self, wideip):
        super(Aaaas, self).__init__(wideip)
        self._meta_data['allowed_lazy_attributes'] = [Aaaa]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:wideip:aaaa:aaaastate': Aaaa}


class Aaaa(Resource):
    """v12.x BIG-IP® AAAA wideip resource"""
    def __init__(self, aaaas):
        super(Aaaa, self).__init__(aaaas)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:wideip:aaaa:aaaastate'


class Cnames(Collection):
    """v12.x BIG-IP® CNAME wideip collection"""
    def __init__(self, wideip):
        super(Cnames, self).__init__(wideip)
        self._meta_data['allowed_lazy_attributes'] = [Cname]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:wideip:cname:cnamestate': Cname}


class Cname(Resource):
    """v12.x BIG-IP® CNAME wideip resource"""
    def __init__(self, cnames):
        super(Cname, self).__init__(cnames)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:wideip:cname:cnamestate'


class Mxs(Collection):
    """v12.x BIG-IP® MX wideip collection."""
    def __init__(self, wideip):
        super(Mxs, self).__init__(wideip)
        self._meta_data['allowed_lazy_attributes'] = [Mx]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:wideip:mx:mxstate': Mx}


class Mx(Resource):
    """v12.x BIG-IP® MX wideip resource"""
    def __init__(self, mxs):
        super(Mx, self).__init__(mxs)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:wideip:mx:mxstate'


class Naptrs(Collection):
    """v12.x BIG-IP® NAPTR wideip collection"""
    def __init__(self, wideip):
        super(Naptrs, self).__init__(wideip)
        self._meta_data['allowed_lazy_attributes'] = [Naptr]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:wideip:naptr:naptrstate': Naptr}


class Naptr(Resource):
    """v12.x BIG-IP® NAPTR wideip resource"""
    def __init__(self, naptrs):
        super(Naptr, self).__init__(naptrs)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:wideip:naptr:naptrstate'


class Srvs(Collection):
    """v12.x BIG-IP® SRV wideip collection"""
    def __init__(self, wideip):
        super(Srvs, self).__init__(wideip)
        self._meta_data['allowed_lazy_attributes'] = [Srv]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:wideip:srv:srvstate': Srv}


class Srv(Resource):
    """v12.x BIG-IP® SRV wideip resource"""
    def __init__(self, naptrs):
        super(Srv, self).__init__(naptrs)
        self._meta_data['required_json_kind'] = \
            'tm:gtm:wideip:srv:srvstate'


class WideipCollection(Collection):
    """v11.x BIG-IP® GTM wideip collection"""
    def __init__(self, gtm):
        self.__class__.__name__ = 'Wideips'
        super(WideipCollection, self).__init__(gtm)
        self._meta_data['allowed_lazy_attributes'] = [Wideip]
        self._meta_data['attribute_registry'] = \
            {'tm:gtm:wideip:wideipstate': Wideip}


class Wideip(Resource):
    """v11.x BIG-IP® GTM pool resource"""
    def __init__(self, wideips):
        super(Wideip, self).__init__(wideips)
        self._meta_data['required_json_kind'] = 'tm:gtm:wideip:wideipstate'



# coding=utf-8
#
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""iWorkflow® Device Groups (shared) module for CM Cloud Managed Devices

REST URI
    ``http://localhost/mgmt/shared/resolver/device-groups/cm-cloud-managed-devices``
"""

from f5.iworkflow.resource import Collection
from f5.iworkflow.resource import OrganizingCollection
from f5.iworkflow.resource import Resource


class Cm_Cloud_Managed_Devices(OrganizingCollection):
    def __init__(self, device_groups):
        super(Cm_Cloud_Managed_Devices, self).__init__(device_groups)
        self._meta_data['required_json_kind'] = \
            'shared:resolver:device-groups:devicegroupstate'
        self._meta_data['attribute_registry'] = {
            'cm:shared:licensing:pools:licensepoolmembercollectionstate':
                Devices_s
        }
        self._meta_data['allowed_lazy_attributes'] = [
            Devices_s
        ]


class Devices_s(Collection):
    def __init__(self, cm_cloud_managed_devices):
        super(Devices_s, self).__init__(cm_cloud_managed_devices)
        self._meta_data['allowed_lazy_attributes'] = [Device]
        self._meta_data['required_json_kind'] = \
            'shared:resolver:device-groups:devicegroupdevicecollectionstate'
        self._meta_data['attribute_registry'] = {
            'shared:resolver:device-groups:restdeviceresolverdevicestate': Device  # NOQA
        }


class Device(Resource):
    def __init__(self, devices_s):
        super(Device, self).__init__(devices_s)
        self._meta_data['required_json_kind'] = \
            'shared:resolver:device-groups:restdeviceresolverdevicestate'
        self._meta_data['required_creation_parameters'] = {
            'address', 'password', 'userName'}
        self._meta_data['required_load_parameters'] = {'uuid', }
